#include "ExpanderOperation.hpp"
#include "ExpanderDAPI.hpp"

ExpanderOperation:: ExpanderOperation(OPERATION_TYPE oprationType, string helpContent) : OperationBaseClass(oprationType, helpContent)
{

}

ExpanderOperation::~ ExpanderOperation()
{

}

int
ExpanderOperation::executeOperation(StringParser commandLine)
{
    string  parameter;
    
    commandLine.getParameter(&parameter, 0);

    if (parameter == HELP)
    {
        showOperationHelp();
        return 0;
    }

    //Execute operation
    switch (getOperationType())
    {
        case OPERATION_CONTROLLER:
        case OPERATION_PORT:
        case OPERATION_PHY:
            return list(commandLine);
        case OPERATION_ALL:
        case OPERATION_ATTACHED_TYPE:
        case OPERATION_LINK:
        case OPERATION_SPEED:
        case OPERATION_ADDRESS:
            return show(commandLine);
        case OPERATION_RESET_CONTROLLER:
            return reset(commandLine);
        case OPERATION_ERROR_ALL:
        case OPERATION_DISPARITY_ERR:
        case OPERATION_INVALID_DWORD_ERR:
        case OPERATION_LOSS_DWORD_SYNC:
        case OPERATION_PHY_RESET_PRBM:
            return displayError(commandLine);
        case OPERATION_CLEAR:
            return clearPHYError(commandLine);
        default:
            cout << UNKNOWN_OPERATION_MSG << endl;
            break;
    }

    return -1;
}

void
ExpanderOperation::showOperationHelp()
{
    string strOpName, strHelp;

    getName(&strOpName);
    getHelp(&strHelp);

    cout << strHelp << endl;
    cout << "Syntax " << strOpName << " ";
    
    switch (getOperationType())
    {
        case OPERATION_ALL:
        case OPERATION_ATTACHED_TYPE:
        case OPERATION_LINK:
        case OPERATION_SPEED:
        case OPERATION_ADDRESS:
            cout << "P1" << endl;
            cout << "P1: Device Name" << endl;
            break;
        case OPERATION_RESET_CONTROLLER:
            cout << "P1" << endl;
            cout << "P1: Controller Name" << endl;
            break;
        case OPERATION_ERROR_ALL:
        case OPERATION_DISPARITY_ERR:
        case OPERATION_INVALID_DWORD_ERR:
        case OPERATION_LOSS_DWORD_SYNC:
        case OPERATION_PHY_RESET_PRBM:
            cout << "P1" << endl;
            cout << "P1: PHY Name" << endl;
            break;
        case OPERATION_CLEAR:
            cout << endl;
            break;
        default:
            break;
    }
}

int
ExpanderOperation::list(StringParser commandLine)
{
    int             parameterNumber = 0;
    ErrorMsg        *pErrorMsg;
    ExpanderDAPI    *pExpanderDAPI = NULL;
    
    commandLine.getParameterNumber(&parameterNumber);
    
    //Check parameter number
    if (parameterNumber != 0)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }
        
    //Creat Expander DAPI object 
    pExpanderDAPI = new ExpanderDAPI("ExpanderDAPI");
    if ((pErrorMsg = pExpanderDAPI->getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    switch (getOperationType())
    {
        case OPERATION_CONTROLLER:
            pExpanderDAPI->listDevice(EXP_CONTROLLER);
            break;
        case OPERATION_PORT:
            pExpanderDAPI->listDevice(EXP_PORT);
            break;
        case OPERATION_PHY:
            pExpanderDAPI->listDevice(EXP_PHY);
            break;
        default:
            break;
    }

    if (pExpanderDAPI != NULL)
    {
        delete pExpanderDAPI;
        pExpanderDAPI = NULL;
    }
    
    return 0;
}

int
ExpanderOperation::show(StringParser commandLine)
{
    int                 parameterNumber = 0;
    string              strDeviceName;
    UINT32              logicalID;
    EXP_DATAELEMENT     dataElement;
    ErrorMsg            *pErrorMsg;
    ExpanderDAPI        *pExpanderDAPI = NULL;

    commandLine.getParameterNumber(&parameterNumber);
    if (parameterNumber != 1)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    //Creat  Expander DAPI object 
    pExpanderDAPI = new ExpanderDAPI("ExpanderDAPI");
    if ((pErrorMsg = pExpanderDAPI->getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    switch (getOperationType())
    {
        case OPERATION_ALL:
            dataElement = EXP_DATAELEMENT_ALL;
            break;
        case OPERATION_ATTACHED_TYPE:
            dataElement = EXP_DATAELEMENT_ATTACHED_TYPE;
            break;
        case OPERATION_LINK:
            dataElement = EXP_DATAELEMENT_LINK;
            break;
        case OPERATION_SPEED:
            dataElement = EXP_DATAELEMENT_SPEED;
            break;
        case OPERATION_ADDRESS:
            dataElement = EXP_DATAELEMENT_ADDRESS;
            break;
        default:
            break;
    }
    
    commandLine.getParameter(&strDeviceName, 0);
    pErrorMsg = pExpanderDAPI->getLogicalIDByName(strDeviceName, &logicalID);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }
    
    pErrorMsg = pExpanderDAPI->showData(logicalID, dataElement);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    if (pExpanderDAPI != NULL)
    {
        delete pExpanderDAPI;
        pExpanderDAPI = NULL;
    }
    
    return 0;
}

int
ExpanderOperation::reset(StringParser commandLine)
{
    int                     parameterNumber = 0;
    UINT32                  logicalID;
    string                  strDeviceName;
    ErrorMsg                *pErrorMsg;
    ExpanderDAPI            *pExpanderDAPI = NULL;

    commandLine.getParameterNumber(&parameterNumber);
    if (parameterNumber != 1)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    //Creat  Expander DAPI object 
    pExpanderDAPI = new ExpanderDAPI("ExpanderDAPI");
    if ((pErrorMsg = pExpanderDAPI->getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    commandLine.getParameter(&strDeviceName, 0);
    pErrorMsg = pExpanderDAPI->getLogicalIDByName(strDeviceName, &logicalID);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    pErrorMsg = pExpanderDAPI->reset(logicalID);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    if (pExpanderDAPI != NULL)
    {
        delete pExpanderDAPI;
        pExpanderDAPI = NULL;
    }
    
    return 0;
}

int
ExpanderOperation::displayError(StringParser commandLine)
{
    int                     parameterNumber = 0;
    UINT32                  logicalID;
    string                  strDeviceName;
    EXP_ERRORELEMENT        errorElement;
    ErrorMsg                *pErrorMsg;
    ExpanderDAPI            *pExpanderDAPI = NULL;

    commandLine.getParameterNumber(&parameterNumber);
    if (parameterNumber != 1)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }
     
    pExpanderDAPI = new ExpanderDAPI("ExpanderDAPI");
    if ((pErrorMsg = pExpanderDAPI->getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    switch (getOperationType())
    {
        case OPERATION_ERROR_ALL:
            errorElement = EXP_ERRORELEMENT_ALL_ERROR;
            break;
        case OPERATION_DISPARITY_ERR:
            errorElement = EXP_ERRORELEMENT_DISPARITY_ERR;
            break;
        case OPERATION_INVALID_DWORD_ERR:
            errorElement = EXP_ERRORELEMENT_INVALID_DWORD_ERR;
            break;
        case OPERATION_LOSS_DWORD_SYNC:
            errorElement = EXP_ERRORELEMENT_LOSS_DWORD_SYNC;
            break;
        case OPERATION_PHY_RESET_PRBM:
            errorElement = EXP_ERRORELEMENT_PHY_RESET_PRBM;
            break;
        default:
            break;
    }
        
    commandLine.getParameter(&strDeviceName, 0);
    pErrorMsg = pExpanderDAPI->getLogicalIDByName(strDeviceName, &logicalID);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    pErrorMsg = pExpanderDAPI->showErrorData(logicalID, errorElement);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    if (pExpanderDAPI != NULL)
    {
        delete pExpanderDAPI;
        pExpanderDAPI = NULL;
    }
    
    return 0;
}

int
ExpanderOperation::clearPHYError(StringParser commandLine)
{
    int                 parameterNumber = 0;
    ErrorMsg            *pErrorMsg;
    ExpanderDAPI        *pExpanderDAPI = NULL;

    //Creat  Expander DAPI object 
    commandLine.getParameterNumber(&parameterNumber);
    if (parameterNumber != 0)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }
     
    pExpanderDAPI = new ExpanderDAPI("ExpanderDAPI");
    if ((pErrorMsg = pExpanderDAPI->getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }
    
    pErrorMsg = pExpanderDAPI->clearPHYError();
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    if (pExpanderDAPI != NULL)
    {
        delete pExpanderDAPI;
        pExpanderDAPI = NULL;
    }
    
    return 0;
}
