#include <sstream>
#include <thread>
#include "MemoryOperation.hpp"
#include "MemoryDAPI.hpp"
#include "ProcessorDAPI.hpp"

MemoryOperation::MemoryOperation(OPERATION_TYPE oprationType, string helpContent) : OperationBaseClass(oprationType, helpContent)
{

}

MemoryOperation::~MemoryOperation()
{
  
}

int
MemoryOperation::executeECCCheck(StringParser in_parser)
{
    int count;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 0)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    return showECCReport();
}

int
MemoryOperation::executeECCClear(StringParser in_parser)
{
    int                     count;
    MemoryDAPI              memoryDAPI("MemoryDAPI");
    ErrorMsg*               pErrorMsg;
    list<ECCInformation>    errorList;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 0)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    if ((pErrorMsg = memoryDAPI.clearECCReport()))
    {
        printError(pErrorMsg);
        return -1;
    }

    return showECCReport();
}

int
MemoryOperation::executeOperation(StringParser in_parser)
{
    string  strParameter;

    in_parser.getParameter(&strParameter, 0);

    if (strParameter == HELP)
    {
        showOperationHelp();
        return 0;
    }

    switch (getOperationType())
    {
        // Operations of Show
        case OPERATION_INFO:
            return executeShowInfo(in_parser);
        case OPERATION_TEST_RANGE:
            return executeShowTestRange(in_parser);
        // Operations of Test
        case OPERATION_ADDR_LINE:
        case OPERATION_DATA_LINE:
        case OPERATION_MARCHC:
        case OPERATION_INJECT:
            return executeTestOnDIMM(in_parser);
        case OPERATION_PATTERN:
        case OPERATION_RANDOM:
            return executeTestOnRange(in_parser);
        case OPERATION_CHECKSUM:
            return executeTestChecksum(in_parser);
        // Operations of Utility
        case OPERATION_WRITE:
            return executeUtilityWrite(in_parser);
        case OPERATION_READ:
            return executeUtilityRead(in_parser);
        // Operation of ECC
        case OPERATION_CHECK:
            return executeECCCheck(in_parser);
        case OPERATION_CLEAR:
            return executeECCClear(in_parser);
        default:
            cout << UNKNOWN_OPERATION_MSG << endl;
            break;
    }

    return -1;
}

int
MemoryOperation::executeShowInfo(StringParser in_parser)
{
    int                 count;
    MemoryDAPI          memoryDAPI("MemoryDAPI");
    ErrorMsg*           pErrorMsg;
    string              strP1;
    DIMMInformation     info;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 1)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    in_parser.getParameter(&strP1, 0);

    // Handle operation
    pErrorMsg = memoryDAPI.getInformation(strP1, &info);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    cout << "Memory Type:           " << info.m_strMemoryType << endl;
    cout << "Module Type:           " << info.m_strModuleType << endl;
    cout << "Speed:                 " << info.m_strMaxSpeed << endl;
    cout << "Size:                  " << info.m_strSize << endl;
    cout << "Rank:                  " << info.m_strRank << endl;
    cout << "Module Manufacturer:   " << info.m_strModuleManufacturer << endl;
    cout << "DRAM Manufacturer:     " << info.m_strDRAMManufacturer << endl;
    cout << "Part Number:           " << info.m_strPartNumber << endl;
    cout << "Serial Number:         " << info.m_strSerialNumber << endl;

    return 0;
}

int
MemoryOperation::executeShowTestRange(StringParser in_parser)
{
    int                 count;
    MemoryDAPI          memoryDAPI("MemoryDAPI");
    ErrorMsg*           pErrorMsg;
    list<MemoryRange>   rangeList, tempRangeList;
    MemoryRange         tempRange;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 0)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    // Handle operation
    pErrorMsg = memoryDAPI.getTestableMemory(&tempRangeList);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    // Merge ranges
    // Do not have to check if there is no range because this case will be considered
    // in error handling
    tempRange = tempRangeList.front();
    tempRangeList.pop_front();

    while (!tempRangeList.empty())
    {
        // We do not have to check if ranges will overlap
        if ((tempRange.m_endAddress + 1) == tempRangeList.front().m_startAddress)
        {
            tempRange.m_endAddress = tempRangeList.front().m_endAddress;
            tempRangeList.pop_front();
        }
        else
        {
            rangeList.push_back(tempRange);
            tempRange = tempRangeList.front();
            tempRangeList.pop_front();
        }
    }

    // Handle the last range
    rangeList.push_back(tempRange);

    cout << "Testable memory range:" << endl;
    while (!rangeList.empty())
    {
        tempRange = rangeList.front();
        cout << "\t" << UINT64ToHexStr(tempRange.m_startAddress);
        cout << " ~ " << UINT64ToHexStr(tempRange.m_endAddress) << endl;

        rangeList.pop_front();
    }

    return 0;
}

int
MemoryOperation::executeTestChecksum(StringParser in_parser)
{
    int                 count = 0;
    string              strP1, strP2, strP3;
    MemoryDAPI          memoryDAPI("MemoryDAPI");
    ErrorMsg*           pErrorMsg = NULL;
    UINT64              u64StartAddress, u64EndAddress, u64Size;

    // Check syntax
    in_parser.getParameterNumber(&count);

    if (count != 3)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);
    in_parser.getParameter(&strP2, 1);
    in_parser.getParameter(&strP3, 2);

    if (in_parser.isHEXString(strP1) ||
        in_parser.isHEXString(strP2) ||
        in_parser.isHEXString(strP3) ||
        (strP3.size() > 4))
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    u64StartAddress = std::stoull(strP1, NULL, 16);
    u64Size = std::stoull(strP2, NULL, 16);
    u64EndAddress = u64StartAddress + u64Size;

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    // Handle operation
    pErrorMsg = memoryDAPI.checksumTest(u64StartAddress, u64EndAddress, strP3, CHECKSUM_8BITS);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    return 0;
}

int
MemoryOperation::executeTestOnDIMM(StringParser in_parser)
{
    int                 error = 0;
    int                 count = 0;
    MemoryDAPI          memoryDAPI("MemoryDAPI");
    ErrorMsg*           pErrorMsg = NULL;
    list<MemoryRange>   rangeList;
    MemoryRange         tempRange, testRange;
    MemoryTestInfo*     pNewTestInfo;
    list<MemoryTestInfo *> testList;
    list<INT32>         tempChannelList, tempRankList;
    INT32               s32Controller, s32Channel, s32Rank;
    UINT64              u64SizeForController, u64SizeForChannel, u64SizeForRank;

    // Check syntax
    in_parser.getParameterNumber(&count);

    switch (getOperationType())
    {
        case OPERATION_ADDR_LINE:
        case OPERATION_DATA_LINE:
        case OPERATION_MARCHC:
        case OPERATION_INJECT:
            if (count != 0)
            {
                error = -1;
            }
            break;
        default:
            error = -1;
            break;
    }

    if (error)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return error;
    }

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    // Handle operation
    rangeList.clear();
    testList.clear();

    pErrorMsg = memoryDAPI.getTestableMemory(&rangeList);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    try
    {
        // Handle each range
        while (!rangeList.empty())
        {
            tempRange = rangeList.front();
            testRange = tempRange;
            rangeList.pop_front();

            if (tempRange.isInterleave())
            {
                // FIXME: MarchC is not ready for interleave mode
                if (getOperationType() == OPERATION_MARCHC)
                {
                    printf("Error: Not support MarchC when memory is interleaving\n");
                    return -1;
                }
                if (getOperationType() == OPERATION_ADDR_LINE)
                {
                     u64SizeForController = tempRange.getSize() / tempRange.m_controllerList.size();
                }

                // Consider controller
                while (tempRange.m_controllerList.size() > 0)
                {
                    s32Controller = tempRange.m_controllerList.front();
                    tempRange.m_controllerList.pop_front();

                    // Consider channel
                    if (tempRange.m_channelList.size() > 0)
                    {
                        tempChannelList = tempRange.m_channelList.front();
                        tempRange.m_channelList.pop_front();

                        if (getOperationType() == OPERATION_ADDR_LINE)
                        {
                            u64SizeForChannel = u64SizeForController / tempChannelList.size();
                        }

                        while (tempChannelList.size() > 0)
                        {
                            s32Channel = tempChannelList.front();
                            tempChannelList.pop_front();

                            // Consider rank
                            tempRankList = tempRange.m_rankList.front();
                            tempRange.m_rankList.pop_front();

                            if (getOperationType() == OPERATION_ADDR_LINE)
                            {
                                u64SizeForRank = u64SizeForChannel / tempRankList.size();
                            }

                            while (tempRankList.size() > 0)
                            {
                                s32Rank = tempRankList.front();
                                tempRankList.pop_front();

                                pNewTestInfo = new MemoryTestInfo();
                                pNewTestInfo->m_testRange = testRange;
                                pNewTestInfo->m_s32TargetController = s32Controller;
                                pNewTestInfo->m_s32TargetChannel = s32Channel;
                                pNewTestInfo->m_s32TargetRank = s32Rank;
                                switch (getOperationType())
                                {
                                    case OPERATION_ADDR_LINE:
                                        pNewTestInfo->m_s32DataSize = 8;
                                        pNewTestInfo->m_u64TestPattern = 0xFFFFFFFFFFFFFFFFULL;
                                        pNewTestInfo->m_u64TargetTestSize = u64SizeForRank;
                                        break;
                                    case OPERATION_DATA_LINE:
                                        pNewTestInfo->m_s32DataSize = 64;
                                        pNewTestInfo->m_u64TargetTestSize = 8;
                                        break;
                                    case OPERATION_MARCHC:
                                        pNewTestInfo->m_s32DataSize = 64;
                                        pNewTestInfo->m_u64TestPattern = 0xFFFFFFFFFFFFFFFFULL;
                                        pNewTestInfo->m_u64TargetTestSize = 0ULL;
                                        break;
                                    case OPERATION_INJECT:
                                        pNewTestInfo->m_u64TargetTestSize = 8;
                                        break;
                                    default:
                                        break;
                                }

                                testList.push_back(pNewTestInfo);
                            }
                        }
                    } // End of Consider channel

                } // End of Consider controller
            }
            else
            {
                pNewTestInfo = new MemoryTestInfo();
                switch (getOperationType())
                {
                    case OPERATION_ADDR_LINE:
                        pNewTestInfo->m_testRange = testRange;
                        pNewTestInfo->m_s32DataSize = 8;
                        pNewTestInfo->m_u64TestPattern = 0xFFFFFFFFFFFFFFFFULL;
                        break;
                    case OPERATION_DATA_LINE:
                        pNewTestInfo->m_s32DataSize = 64;
                        pNewTestInfo->m_testRange.m_startAddress = testRange.m_startAddress;
                        pNewTestInfo->m_testRange.m_endAddress = testRange.m_startAddress + 7;
                        break;
                    case OPERATION_MARCHC:
                        pNewTestInfo->m_testRange = testRange;
                        pNewTestInfo->m_s32DataSize = 64;
                        pNewTestInfo->m_u64TestPattern = 0xFFFFFFFFFFFFFFFFULL;
                        break;
                    case OPERATION_INJECT:
                        // Testing unit of ECC injection is 64 bits
                        pNewTestInfo->m_testRange.m_startAddress = testRange.m_startAddress;
                        pNewTestInfo->m_testRange.m_endAddress = pNewTestInfo->m_testRange.m_startAddress + 7;
                        break;
                    default:
                        // This error case had been handled in previous steps
                        break;
                }
            }
        } // End of Handle each range

        if (getOperationType() == OPERATION_INJECT)
        {
            INT32               i, s32CPUs = 2, s32Reg = 0x287; // 0x287 for HA, 0x288 for HA1
            list<UINT64>        saveMSRList;
            list<UINT64>::iterator itUINT64;
            CHAR                szTemp[32];
            ProcessorDAPI       processorDAPI("ProcessorDAPI");

            if ((pErrorMsg = processorDAPI.getErrorMsg()) == NO_ERROR_MSG)
            {
                // Save MCE bank7 and bank8
                for (i = 0; i < s32CPUs; i++)
                {
                    sprintf(szTemp, "processor%d", i);
                    if ((pErrorMsg = processorDAPI.Read(szTemp, s32Reg, 2, &saveMSRList)))
                        break;
                }
            }

            if (pErrorMsg == NO_ERROR_MSG)
            {
                // Disable MCE bank7 and bank8 reports
                itUINT64 = saveMSRList.begin();
                for (i = 0; i < s32CPUs; i++, itUINT64++)
                {
                    sprintf(szTemp, "processor%d", i);
                    if ((pErrorMsg = processorDAPI.Write(szTemp, s32Reg, *(itUINT64) & 0xFFFFFFFFBFFFFFFF)))
                        break;

                    itUINT64++;
                    if ((pErrorMsg = processorDAPI.Write(szTemp, s32Reg + 1, *(itUINT64) & 0xFFFFFFFFBFFFFFFF)))
                        break;
                }
            }

            if (pErrorMsg == NO_ERROR_MSG)
                error = executeTestByThreads(testList);

            // Restore MCE bank7 and bank8
            if (pErrorMsg == NO_ERROR_MSG)
            {
                itUINT64 = saveMSRList.begin();
                for (i = 0; i < s32CPUs; i++, itUINT64++)
                {
                    sprintf(szTemp, "processor%d", i);
                    if ((pErrorMsg = processorDAPI.Write(szTemp, s32Reg, *(itUINT64))))
                        break;

                    itUINT64++;
                    if ((pErrorMsg = processorDAPI.Write(szTemp, s32Reg + 1, *(itUINT64))))
                        break;
                }
            }

            if (pErrorMsg)
            {
                error = -1;
                printError(pErrorMsg);
            }
        }
        else
        {
            error = executeTestByThreads(testList);
        }
    }
    catch (exception& e)
    {
        cout << "Exception: executeTestCommon - " << e.what() << endl;
        error = -1;
    }

    while (!testList.empty())
    {
        if (testList.front())
        {
            if (testList.front()->m_pErrorMsg)
            {
                printError(testList.front()->m_pErrorMsg);
                error = -1;
            }
            delete testList.front();
        }

        testList.pop_front();
    }

    return error;
}

int
MemoryOperation::executeTestOnRange(StringParser in_parser)
{
    int                 error = 0;
    int                 count = 0;
    string              strP1, strP2, strP3;
    MemoryDAPI          memoryDAPI("MemoryDAPI");
    ErrorMsg*           pErrorMsg = NULL;
    UINT64              u64StartAddress, u64EndAddress, u64PageSize, u64BlockSize;
    list<MemoryRange>   tempList, rangeList;
    MemoryRange         tempRange, testRange;
    MemoryTestInfo*     pNewTestInfo;
    list<MemoryTestInfo *> testList;
    UINT64              u64CPUNum, u64Count;

    // Check syntax
    in_parser.getParameterNumber(&count);

    switch (getOperationType())
    {
        case OPERATION_PATTERN:
            if (count != 2)
            {
                error = -1;
            }
            else
            {
                in_parser.getParameter(&strP1, 0);
                in_parser.getParameter(&strP2, 1);

                if (in_parser.isHEXString(strP1) ||
                    in_parser.isHEXString(strP2))
                {
                    error = -1;
                }
            }
            break;
        case OPERATION_RANDOM:
            if (count != 3)
            {
                error = -1;
            }
            else
            {
                in_parser.getParameter(&strP1, 0);
                in_parser.getParameter(&strP2, 1);
                in_parser.getParameter(&strP3, 2);

                if (in_parser.isHEXString(strP1) || in_parser.isHEXString(strP2))
                {
                    error = -1;
                }
            }
            break;
        default:
            error = -1;
            break;
    }

    if (error)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return error;
    }

    u64StartAddress = std::stoull(strP1, NULL, 16);
    u64EndAddress = std::stoull(strP2, NULL, 16);

    if (u64EndAddress <= u64StartAddress)
    {
        cout << "Error: End address should be higher than start address" << endl;
        return -1;
    }

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    // Handle operation
    rangeList.clear();
    testList.clear();

    pErrorMsg = memoryDAPI.getTestableMemory(&rangeList, u64StartAddress, u64EndAddress);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    tempRange = rangeList.front();
    rangeList.pop_front();

    while (!rangeList.empty())
    {
        // We do not have to check if ranges will overlap
        if ((tempRange.m_endAddress + 1) == rangeList.front().m_startAddress)
        {
            tempRange.m_endAddress = rangeList.front().m_endAddress;
            rangeList.pop_front();
        }
        else
        {
            tempList.push_back(tempRange);
            tempRange = rangeList.front();
            rangeList.pop_front();
        }
    }

    // Handle the last range
    tempList.push_back(tempRange);

    u64PageSize = getpagesize();

    while (!tempList.empty())
    {
        if (tempList.front().m_startAddress & (u64PageSize - 1))
        {
            tempRange = tempList.front();
            tempRange.m_startAddress &= ~(u64PageSize - 1);
            tempRange.m_startAddress += u64PageSize;

            if (tempRange.m_endAddress <= tempRange.m_startAddress)
            {
                tempList.pop_front();
                continue;
            }
        }
        else
            tempRange = tempList.front();

        tempList.pop_front();

        u64BlockSize = (UINT64)memoryDAPI.getTestBlockSize();
        u64Count = tempRange.getSize() / u64BlockSize;

        u64CPUNum = sysconf(_SC_NPROCESSORS_ONLN);
        u64CPUNum--;
        if (u64Count >= u64CPUNum)
        {
            if (u64Count % u64CPUNum)
            {
                u64Count /= --u64CPUNum;                                                             
            }
            else
            {
                u64Count /= u64CPUNum;
            }
            u64BlockSize *= u64Count;
        }
        testRange = tempRange;

        u64EndAddress = testRange.m_startAddress + u64BlockSize - 1;
        while(u64EndAddress <= tempRange.m_endAddress)
        {
            testRange.m_endAddress = u64EndAddress;
            rangeList.push_back(testRange);
            testRange.m_startAddress = u64EndAddress + 1;
            u64EndAddress += u64BlockSize;
        }

        // Handle the last block smaller than the test block
        if (testRange.m_endAddress < tempRange.m_endAddress)
        {
            testRange.m_startAddress = testRange.m_endAddress + 1;
            testRange.m_endAddress = tempRange.m_endAddress;
            rangeList.push_back(testRange);
        }
    }

    try
    {
        while (!rangeList.empty())
        {
            // Create testing information to run tests by threads
            pNewTestInfo = new MemoryTestInfo();

            switch (getOperationType())
            {
                case OPERATION_PATTERN:
                    pNewTestInfo->m_testRange = rangeList.front();
                    break;
                case OPERATION_RANDOM:
                    pNewTestInfo->m_testRange = rangeList.front();
                    pNewTestInfo->m_u32TestTime = std::stoul(strP3, NULL, 10);
                    break;
                default:
                    // This error case had been handled in previous steps
                    break;
            }

            // Adjust end address for DAPIs
            pNewTestInfo->m_testRange.m_endAddress++;

            testList.push_back(pNewTestInfo);

            rangeList.pop_front();
        }

        error = executeTestByThreads(testList);

    }
    catch (exception& e)
    {
        cout << "Exception: executeTestCommon - " << e.what() << endl;
        error = -1;
    }

    while (!testList.empty())
    {
        if (testList.front())
        {
            if (testList.front()->m_pErrorMsg)
            {
                printError(testList.front()->m_pErrorMsg);
                error = -1;
            }
            delete testList.front();
        }

        testList.pop_front();
    }

    return error;
}

int
MemoryOperation::executeUtilityWrite(StringParser in_parser)
{
    int         count;
    string      strP1, strP2, strP3;
    MemoryDAPI  memoryDAPI("MemoryDAPI");
    ErrorMsg*   pErrorMsg;
    UINT64      u64Address;
    UINT32      u32Unit;
    INT32       s32CharacterUnit = 0, s32Offset = 0, s32HandledSize = 0;
    string      strTemp;
    UINT8       u8Data;
    UINT16      u16Data;
    UINT32      u32Data;
    UINT64      u64Data;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 3)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);
    in_parser.getParameter(&strP2, 1);
    in_parser.getParameter(&strP3, 2);

    if (in_parser.isHEXString(strP1) ||
        isValidMemoryUnit(strP2) ||
        in_parser.isHEXString(strP3))
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    // Handle operation
    u64Address = std::stoull(strP1, NULL, 16);
    u32Unit = std::stoul(strP2);
    s32CharacterUnit = u32Unit >> 2;
    // Remove 0x from input data
    strP3.erase(0, 2);
    if (strP3.size() % s32CharacterUnit)
    {
        strP3.insert(0, (s32CharacterUnit - strP3.size() % s32CharacterUnit), '0');
    }
    s32Offset = strP3.size();

    s32HandledSize = 0;

    while (s32CharacterUnit <= s32Offset)
    {
        // Get current data from input string
        s32Offset -= s32CharacterUnit;
        strTemp.clear();
        strTemp = strP3.substr(s32Offset, s32CharacterUnit);

        switch (u32Unit)
        {
            case 8:
                u8Data = (UINT8)std::stoul(strTemp, NULL, 16);
                pErrorMsg = memoryDAPI.writeMemory8(u64Address, u8Data);
                u64Address++;
                s32HandledSize++;
                break;
            case 16:
                u16Data = (UINT16)std::stoul(strTemp, NULL, 16);
                pErrorMsg = memoryDAPI.writeMemory16(u64Address, u16Data);
                u64Address += 2;
                s32HandledSize += 2;
                break;
            case 32:
                u32Data = (UINT32)std::stoul(strTemp, NULL, 16);
                pErrorMsg = memoryDAPI.writeMemory32(u64Address, u32Data);
                u64Address += 4;
                s32HandledSize += 4;
                break;
            case 64:
                u64Data = (UINT64)std::stoul(strTemp, NULL, 16);
                pErrorMsg = memoryDAPI.writeMemory64(u64Address, u64Data);
                u64Address += 8;
                s32HandledSize += 8;
                break;
            default:
                // This case will not be happened
                break;
        }

        if (pErrorMsg)
        {
            u64Address = std::stoull(strP1, NULL, 16);
            s32HandledSize -= (u32Unit >> 3);
            showMemoryContent(u64Address, u32Unit, (UINT64)s32HandledSize);
            printError(pErrorMsg);
            return -1;
        }
    }

    u64Address = std::stoull(strP1, NULL, 16);
    return showMemoryContent(u64Address, u32Unit, (UINT64)s32HandledSize);
}

int
MemoryOperation::executeUtilityRead(StringParser in_parser)
{
    int     count;
    string  strP1, strP2, strP3;
    UINT64  u64Address;
    UINT32  u32Unit;
    UINT64  u64Size;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 3)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);
    in_parser.getParameter(&strP2, 1);
    in_parser.getParameter(&strP3, 2);

    if (in_parser.isHEXString(strP1) ||
        isValidMemoryUnit(strP2) ||
        in_parser.isHEXString(strP3))
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    // Handle operation
    u64Address = std::stoull(strP1, NULL, 16);
    u32Unit = std::stoul(strP2);
    u64Size = std::stoull(strP3, NULL, 16);

    return showMemoryContent(u64Address, u32Unit, u64Size);
}


int
MemoryOperation::executeTestByThreads(list<MemoryTestInfo *> in_testInfoList)
{
    int                 error = 0;
    BOOL                bHasAvailableCPU = FALSE;
    list<thread *>      threadList;
    thread*             pProgressHandler = NULL;

    list<MemoryTestInfo *>::iterator    itTestInfo;
    list<MemoryDAPI *>::iterator        itDAPI;

    m_s32AvailableCPUAmount = sysconf(_SC_NPROCESSORS_ONLN);

    try {
        if (m_s32AvailableCPUAmount > 1)
        {
            m_s32AvailableCPUAmount--;
            pProgressHandler = new thread(&MemoryOperation::handleTestProgress, this, in_testInfoList);
        }

        // Create thread for each test when there is an available CPU
        while (!in_testInfoList.empty())
        {
            bHasAvailableCPU = FALSE;

            m_mutex.lock();
            if (m_s32AvailableCPUAmount > 0)
            {
                bHasAvailableCPU = TRUE;
                m_s32AvailableCPUAmount--;
            }
            m_mutex.unlock();

            if (bHasAvailableCPU)
            {
                threadList.push_back(new thread(&MemoryOperation::executeTestInThread, this, in_testInfoList.front()));
                in_testInfoList.pop_front();
            }
        }
    }
    catch (exception& e)
    {
        cout << "Exception: executeTestByThreads - " << e.what() << endl;
        error = -1;
    }

    while (!threadList.empty())
    {
        if (threadList.front())
        {
            threadList.front()->join();

            delete threadList.front();
        }

        threadList.pop_front();
    }

    if (pProgressHandler)
    {
        pProgressHandler->join();
        delete pProgressHandler;
    }

    return error;
}

void
MemoryOperation::executeTestInThread(MemoryTestInfo* in_pTestInfo)
{
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    MemoryDAPI* pMemoryDAPI;

    try
    {
        pMemoryDAPI = new MemoryDAPI("MemoryDAPI");

        in_pTestInfo->m_pMemoryDAPI = pMemoryDAPI;

        pErrorMsg = pMemoryDAPI->getErrorMsg();
        if (pErrorMsg == NO_ERROR_MSG)
        {
            // Execute test
            switch (getOperationType())
            {
                case OPERATION_ADDR_LINE:
                    if (in_pTestInfo->m_testRange.isInterleave())
                    {
                        pErrorMsg = pMemoryDAPI->addressLineTest(in_pTestInfo->m_testRange.m_startAddress,
                                                                 in_pTestInfo->m_u64TargetTestSize,
                                                                 in_pTestInfo->m_s32TargetController,
                                                                 in_pTestInfo->m_s32TargetChannel,
                                                                 in_pTestInfo->m_s32TargetRank,
                                                                 in_pTestInfo->m_s32DataSize,
                                                                 in_pTestInfo->m_u64TestPattern);
                    }
                    else
                    {
                        pErrorMsg = pMemoryDAPI->addressLineTest(in_pTestInfo->m_testRange.m_startAddress,
                                                                 in_pTestInfo->m_testRange.m_endAddress,
                                                                 in_pTestInfo->m_s32DataSize,
                                                                 in_pTestInfo->m_u64TestPattern);
                    }
                    break;
                case OPERATION_DATA_LINE:
                    if (in_pTestInfo->m_testRange.isInterleave())
                    {
                        pErrorMsg = pMemoryDAPI->dataLineTest(in_pTestInfo->m_testRange.m_startAddress,
                                                              in_pTestInfo->m_u64TargetTestSize,
                                                              in_pTestInfo->m_s32TargetController,
                                                              in_pTestInfo->m_s32TargetChannel,
                                                              in_pTestInfo->m_s32TargetRank,
                                                              in_pTestInfo->m_s32DataSize);
                    }
                    else
                    {
                        pErrorMsg = pMemoryDAPI->dataLineTest(in_pTestInfo->m_testRange.m_startAddress,
                                                              in_pTestInfo->m_testRange.m_endAddress,
                                                              in_pTestInfo->m_s32DataSize);
                    }

                    break;
                case OPERATION_MARCHC:
                    pErrorMsg = pMemoryDAPI->marchCTest(in_pTestInfo->m_testRange.m_startAddress,
                                                        in_pTestInfo->m_testRange.m_endAddress,
                                                        in_pTestInfo->m_s32DataSize,
                                                        in_pTestInfo->m_u64TestPattern);
                    break;
                case OPERATION_PATTERN:
                    pErrorMsg = pMemoryDAPI->patternTest(in_pTestInfo->m_testRange.m_startAddress,
                                                         in_pTestInfo->m_testRange.m_endAddress);
                    break;
                case OPERATION_RANDOM:
                    pErrorMsg = pMemoryDAPI->randomAddressTest(in_pTestInfo->m_testRange.m_startAddress,
                                                               in_pTestInfo->m_testRange.m_endAddress,
                                                               in_pTestInfo->m_u32TestTime);
                    break;
                case OPERATION_INJECT:
                    if (in_pTestInfo->m_testRange.isInterleave())
                    {
                        pErrorMsg = pMemoryDAPI->eccErrorDetectionTest(in_pTestInfo->m_testRange.m_startAddress,
                                                                       in_pTestInfo->m_u64TargetTestSize,
                                                                       in_pTestInfo->m_s32TargetController,
                                                                       in_pTestInfo->m_s32TargetChannel,
                                                                       in_pTestInfo->m_s32TargetRank);
                    }
                    else
                    {
                        pErrorMsg = pMemoryDAPI->eccErrorDetectionTest(in_pTestInfo->m_testRange.m_startAddress,
                                                                       in_pTestInfo->m_testRange.m_endAddress);
                    }
                    break;
                default:
                    break;
            }
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            in_pTestInfo->m_pErrorMsg = pErrorMsg;
        }
    }
    catch (exception& e)
    {
        cout << "Exception: executeTestInThread - " << e.what() << endl;
    }

    in_pTestInfo->m_bIsTestDone = TRUE;

    m_mutex.lock();
    m_s32AvailableCPUAmount++;
    m_mutex.unlock();
}

void
MemoryOperation::handleTestProgress(list<MemoryTestInfo *> in_testInfoList)
{
    INT32   s32Progress = 0;
    INT32   s32Divisor = in_testInfoList.size();

    list<MemoryTestInfo *>::iterator    itTestInfo;

    // Delay to wait DAPI to initialize its progress
    usleep(1000);

    while (!in_testInfoList.empty())
    {
        s32Progress = 0;

        if ((INT32)in_testInfoList.size() < s32Divisor)
        {
            s32Progress = (s32Divisor - in_testInfoList.size()) * 100;
        }

        itTestInfo = in_testInfoList.begin();

        while (itTestInfo != in_testInfoList.end())
        {
            if ((*itTestInfo)->m_pMemoryDAPI)
            {
                if ((*itTestInfo)->m_bIsTestDone)
                {
                    s32Progress += 100;

                    delete (*itTestInfo)->m_pMemoryDAPI;

                    (*itTestInfo)->m_pMemoryDAPI = NULL;

                    itTestInfo = in_testInfoList.erase(itTestInfo);
                }
                else
                {
                    s32Progress += (*itTestInfo)->m_pMemoryDAPI->getProgress();

                    itTestInfo++;
                }
            }
            else
            {
                itTestInfo++;
            }
        }
        s32Progress /= s32Divisor;

        printf("Progressing:%3d%%\r", s32Progress);
        fflush(stdout);

        usleep(1000000);
    }

    printf("Progressing:%3d%%\r", 100);
    fflush(stdout);

    m_mutex.lock();
    if (sysconf(_SC_NPROCESSORS_CONF) > 1)
    {
        m_s32AvailableCPUAmount++;
    }
    m_mutex.unlock();

    printf("\n");
}

int
MemoryOperation::isValidMemoryUnit(string in_strUnit)
{
    if ((in_strUnit == "8") || (in_strUnit == "16") || (in_strUnit == "32") || (in_strUnit == "64"))
        return 0;
    else
        return -1;
}

int
MemoryOperation::showECCReport(void)
{
    MemoryDAPI              memoryDAPI("MemoryDAPI");
    ErrorMsg*               pErrorMsg;
    list<ECCInformation>    errorList;
    list<ECCInformation>::iterator itTempError;
    INT32 s32TotalCount = 0;

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    if ((pErrorMsg = memoryDAPI.getECCReport(&errorList)))
    {
        printError(pErrorMsg);
        return -1;
    }

    itTempError = errorList.begin();
    while (itTempError != errorList.end())
    {
        if (itTempError->m_strDIMMName.empty())
        {
            itTempError->m_strDIMMName = "unknown";
        }

        s32TotalCount += itTempError->m_s32ErrorCount;

        itTempError++;
    }

    cout << "ECC error counts:" << to_string(s32TotalCount) << endl;

    itTempError = errorList.begin();
    while (itTempError != errorList.end())
    {
        cout << "  " << itTempError->m_strDIMMName;
        cout << ":" << to_string(itTempError->m_s32ErrorCount) << endl;
        itTempError++;
    }

    return 0;
}

int
MemoryOperation::showMemoryContent(UINT64 in_u64Address,
                                   UINT32 in_u32Unit,
                                   UINT64 in_u64Size)
{
    MemoryDAPI  memoryDAPI("MemoryDAPI");
    ErrorMsg*   pErrorMsg;
    UINT32      u32UnitByteCount;
    INT32       s32TempCount;
    std::stringstream stream;

    if ((pErrorMsg = memoryDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    u32UnitByteCount = in_u32Unit >> 3;

    if (in_u64Size % u32UnitByteCount)
    {
        in_u64Size >>= u32UnitByteCount;
        in_u64Size <<= u32UnitByteCount;
    }

    while (in_u64Size > 0)
    {
        // Show start address
        cout << UINT64ToHexStr(in_u64Address) << ":";

        // Show data
        if (in_u64Size >> 4)
        {
            // This line has 16 bytes
            s32TempCount = 16 / u32UnitByteCount;
            in_u64Size -= 16;
        }
        else
        {
            // Handle left bytes for this line
            s32TempCount = in_u64Size / u32UnitByteCount;
            in_u64Size = 0;
        }

        // Handle one line
        stream.str("");
        stream.clear();
        while (s32TempCount--)
        {
            stream << " ";

            switch (in_u32Unit)
            {
                case 8:
                    stream << std::hex << std::setw(2) << std::setfill('0') << (UINT16)memoryDAPI.readMemory8(in_u64Address++);
                    break;
                case 16:
                    stream << std::hex << std::setw(4) << std::setfill('0') << memoryDAPI.readMemory16(in_u64Address);
                    in_u64Address += u32UnitByteCount;
                    break;
                case 32:
                    stream << std::hex << std::setw(8) << std::setfill('0') << memoryDAPI.readMemory32(in_u64Address);
                    in_u64Address += u32UnitByteCount;
                    break;
                case 64:
                    stream << std::hex << std::setw(16) << std::setfill('0') << memoryDAPI.readMemory64(in_u64Address);
                    in_u64Address += u32UnitByteCount;
                    break;
                default:
                    break;
            }
        }

        cout << stream.str() << endl;
    }

    return 0;
}

void
MemoryOperation::showOperationHelp()
{
    string strOpName, strHelp;

    getName(&strOpName);
    getHelp(&strHelp);

    cout << strHelp << endl;
    cout << "Syntax " << strOpName << " ";

    switch (getOperationType())
    {
        // Operations of Show
        case OPERATION_INFO:
            cout << "P1" << endl;
            cout << "P1: device name" << endl;
            break;
        case OPERATION_TEST_RANGE:
            cout << endl;
            break;
        // Operations of Test
        case OPERATION_ADDR_LINE:
            cout << endl;
            break;
        case OPERATION_DATA_LINE:
            cout << endl;
            break;
        case OPERATION_MARCHC:
            cout << endl;
            break;
        case OPERATION_PATTERN:
            cout << "P1 P2" << endl;
            cout << "P1: start address (Hex)" << endl;
            cout << "P2: end address (Hex)" << endl;
            break;
        case OPERATION_RANDOM:
            cout << "P1 P2 P3" << endl;
            cout << "P1: start address (Hex)" << endl;
            cout << "P2: end address (Hex)" << endl;
            cout << "P3: execution time in second (Dec)" << endl;
            break;
        case OPERATION_CHECKSUM:
            cout << "P1 P2 P3" << endl;
            cout << "P1: start address (Hex)" << endl;
            cout << "P2: size (Hex)" << endl;
            cout << "P3: expected 8 bits checksum value (Hex)" << endl;
            break;
        // Operations of Utility
        case OPERATION_WRITE:
            cout << "P1 P2 P3" << endl;
            cout << "P1: start address (Hex)" << endl;
            cout << "P2: unit (8/16/32/64)" << endl;
            cout << "P3: data (Hex)" << endl;
            break;
        case OPERATION_READ:
            cout << "P1 P2 P3" << endl;
            cout << "P1: start address (Hex)" << endl;
            cout << "P2: unit (8/16/32/64)" << endl;
            cout << "P3: size (Hex)" << endl;
            break;
        // Operations of ECC
        case OPERATION_CHECK:
            cout << endl;
            break;
        case OPERATION_CLEAR:
            cout << endl;
            break;
        case OPERATION_INJECT:
            cout << endl;
            break;
        default:
            break;
    }
}
