#include "NVSRAMOperation.hpp"

NVSRAMOperation::NVSRAMOperation(OPERATION_TYPE OprationType, string helpContent) : OperationBaseClass(OprationType, helpContent)
{
    nvsramDAPI = NULL;
    memDAPI = NULL;

    switch(OprationType)
    {
        case OPERATION_TEST_RANGE:
            setName("testrange");
            setHelp(helpContent);
            operationType = OprationType;

            break;
        case OPERATION_DATALINE:
            setName("dataline");
            setHelp(helpContent);
            operationType = OprationType;

            break;
        case OPERATION_ADDRLINE:
            setName("addrline");
            setHelp(helpContent);
            operationType = OprationType;

            break;
        case OPERATION_NVSTEST:
            setName("nvstest");
            setHelp(helpContent);
            operationType = OprationType;

            break;
        default:
            break;
    }
}

NVSRAMOperation::~NVSRAMOperation()
{
    if(nvsramDAPI != NULL)
        delete nvsramDAPI;
}

int NVSRAMOperation::executeOperation(StringParser CommandLine)
{
    string parameter1;
    COMMAND_TYPE CommandType;
    int err = 0;

    CommandLine.getParameter(&parameter1, 0);
    CommandLine.getCommandType(&CommandType);

    //If the first parameter is "help" or space, show help message
    if(CommandType == CMD_LIST || CommandType == CMD_SHOW)
    {
        parameter1 = NO_OPRATION;
    }
    else
    {
        if(parameter1 == HELP || parameter1 == "")
        {
            helpParameter();
            err = -1;
            return err;
        }
    }

    //Execute operation
    switch(operationType)
    {
        case OPERATION_TEST_RANGE:
            err = executeShow(CommandLine);

            break;
        case OPERATION_DATALINE:
        case OPERATION_ADDRLINE:
        case OPERATION_NVSTEST:
            err = executeTest(CommandLine);

            break;
        default:
            cout << "Unsupport operation" << endl;
            err = -1;

            break;
    }

    return err;
}

int NVSRAMOperation::helpParameter()
{
    //Show parameter help message depend on operation type
    switch(operationType)
    {
        case OPERATION_TEST_RANGE:
            cout << "Show NVSRAM device test range" << endl;
            cout << "Syntax: testrange" << endl;

            break;
        case OPERATION_ADDRLINE:
            cout << "NVSRAM Device Data Line Test" << endl;
            cout << "Syntax: addrline P1 P2" << endl;
            cout << "P1: Start Address(Hex)" << endl;
            cout << "P2: End Address(Hex)" << endl;

            break;
        case OPERATION_DATALINE:
            cout << "NVSRAM Device Address Line Test" << endl;
            cout << "Syntax: dataline P1 P2" << endl;
            cout << "P1: Start Address(Hex)" << endl;
            cout << "P2: End Address(Hex)" << endl;

            break;
        case OPERATION_NVSTEST:
            cout << "NVSRAM DeviceNon-volatile Test" << endl;
            cout << "Syntax: nvstest P1 P2 P3" << endl;
            cout << "P1: Start Address(Hex)" << endl;
            cout << "P2: End Address(Hex)" << endl;

            break;
        default:
            break;
    }

    return 0;
}

int NVSRAMOperation::executeShow(StringParser CommandLine)
{
    int err = 0;
    vector<NVSRAM_RESOURCE_INFO> nvsramList;
    ErrorMsg* errorMsg = NO_ERROR_MSG;
    string showErrMsg;

    showErrMsg.clear();

    nvsramDAPI = new NVSRAMDAPI("NVSRAMDAPI");

    err = checkSyntax(CommandLine);
    if(err == 0)
    {
        switch(operationType)
        {
            case OPERATION_TEST_RANGE:
                PHYSICAL_ADDRESS_T endAddress;
                errorMsg = nvsramDAPI->getInformationList(&nvsramList);

                if(errorMsg != NO_ERROR_MSG)
                {
                    getErrorMessage(errorMsg, &showErrMsg);
                    cout << showErrMsg.c_str() << endl;
                    return -1;
                }
                // only one NVSRAM on the host currently
                for(vector<NVSRAM_RESOURCE_INFO>::iterator iter = nvsramList.begin(); iter != nvsramList.end(); ++iter)
                {
                    endAddress = iter->BaseAddress + iter->BARSize - 1;
                    cout << std::showbase << std::internal << std::hex << std::setw(18) << std::setfill('0') << iter->BaseAddress << " ~ ";
                    cout << std::showbase << std::internal << std::hex << std::setw(18) << std::setfill('0') << endAddress << endl;
                }

                break;
            default:
                return -1;
        }
    }
    return err;
}

int NVSRAMOperation::executeTest(StringParser CommandLine)
{
    INT32 err = 0;
    vector<NVSRAM_RESOURCE_INFO> nvsramList;
    vector<NVSRAM_RESOURCE_INFO>::iterator iter;
    string targetName;
    string tempString;
    ErrorMsg* errorMsg = NO_ERROR_MSG;
    UINT64 testStartAddress, testEndAddress, testPattern = 0xFFFFFFFFFFFFFFFFULL;
    string showErrMsg;

    nvsramDAPI = new NVSRAMDAPI("NVSRAMDAPI");
    memDAPI = new MemoryDAPI("MEMDAPI");

    if((errorMsg = nvsramDAPI->getInformationList(&nvsramList)) != NO_ERROR_MSG)
    {
        getErrorMessage(errorMsg, &showErrMsg);
        cout << showErrMsg.c_str() << endl;

        if(nvsramDAPI != NULL)
            delete nvsramDAPI;
        if(memDAPI != NULL)
            delete memDAPI;

        return -1;
    }

    err = checkSyntax(CommandLine);
    if (err == 0)
    {
        iter = nvsramList.begin(); // only one NVSRAM now

        CommandLine.getParameter(&tempString, 0);

        testStartAddress = std::stoull(tempString, NULL, 16);

        tempString.clear();

        CommandLine.getParameter(&tempString, 1);

        testEndAddress = std::stoull(tempString, NULL, 16);

        tempString.clear();

        // check input parameter
        if(testEndAddress < testStartAddress)
        {
            cout << "End address should be equal or higher than start address" << endl;
            return -1;
        }

        if((testStartAddress < (UINT64)(iter->BaseAddress)) ||
           (testStartAddress > ((UINT64)((iter->BaseAddress) + (iter->BARSize) - 1))))
        {
            cout << "Start address is out of testable range" << endl;
            return -1;
        }

        if(testEndAddress > (UINT64)(iter->BaseAddress + iter->BARSize - 1))
        {
            cout << "End address is out of testable range" << endl;
            return -1;
        }

        switch(operationType)
        {
            case OPERATION_ADDRLINE:
                errorMsg = memDAPI->addressLineTest(testStartAddress, testEndAddress, iter->elementSize, testPattern);

                break;
            case OPERATION_DATALINE:
                errorMsg = memDAPI->dataLineTest(testStartAddress, testEndAddress, 8);

                break;
            case OPERATION_NVSTEST:
                errorMsg = nvsramDAPI->nonVolatileTest(testStartAddress, testEndAddress, iter->elementSize);

                break;
            default:
                cout << "Unsupport operation" << endl;
                err = -1;
                break;
        }

        if(errorMsg != NO_ERROR_MSG)
        {
            getErrorMessage(errorMsg, &showErrMsg);
            cout <<  showErrMsg.c_str() << endl;
            return -1;
        }
        else
            cout << "Pass" << endl;
    }

    return err;
}

int NVSRAMOperation::checkSyntax(StringParser CommandLine)
{
    int i, intParameterNumber = 0;
    string parameter[7];
    int err = 0;

    CommandLine.getParameterNumber(&intParameterNumber);

    switch(operationType)
    {
        case NO_OPRATION:
            if(intParameterNumber > 0)
            {
                cout << "Not Necessary Operation Parameter" << endl;
                helpParameter();
                return -1;
            }

            break;
        case OPERATION_TEST_RANGE:
            //Check parameter number
            if(intParameterNumber > 0)
            {
                cout << "Not Necessary Operation Parameter" << endl;
                helpParameter();
                return -1;
            }

            break;
        case OPERATION_DATALINE:
        case OPERATION_ADDRLINE:
        case OPERATION_NVSTEST:
            if(intParameterNumber != 2)
            {
                cout << "Parameter numbers isn't 2" << endl;
                helpParameter();
                return -1;
            }
            //Get parameter
            for(i = 0; i < intParameterNumber; i++)
            {
                CommandLine.getParameter(&parameter[i], i);
            }
            //Check parameter 1 is numeric
            err = CommandLine.isHEXString(parameter[0]);
            if(err != 0)
            {
                cout << "parameter 1 isn't numeric." << endl;
                helpParameter();
                return -1;
            }
            //Check parameter 2 is numeric
            err = CommandLine.isHEXString(parameter[1]);
            if(err != 0)
            {
                cout << "parameter 2 isn't numeric." << endl;
                helpParameter();
                return -1;
            }

            break;
        default:

            break;
    }

    return 0;
}
