/// @file       FPGADAPI.hpp
/// @brief      FPGADAPI class
/// @details    Provide diagnostic and utility APIs for FPGA module
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#include <unistd.h>
#include <stdlib.h>

#include "FPGADAPI.hpp"
#include "MemoryDDAPI.hpp"
#include "CommonFunctions.hpp"

/// @fn         FPGADAPI::FPGADAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
/// @brief      FPGADAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
FPGADAPI::FPGADAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
{
    ErrorMsg *pErrorMsg = NO_ERROR_MSG;

    if ((m_pFPGADDAPI = new FPGADDAPI()) == NULL)
    {
        addErrorMsg(ERROR_FPGA_DAPI_CONSTRUCTOR, "Fail to create FPGADDAPI");
        return;
    }

    if ((pErrorMsg = m_pFPGADDAPI->getErrorMsg()) != NO_ERROR_MSG)
    {
        addErrorMsg(ERROR_FPGA_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
    }
}

/// @fn         FPGADAPI::~FPGADAPI(void)
/// @brief      FPGADAPI destructor
/// @exception  Nothing
/// @return     Nothing
FPGADAPI::~FPGADAPI(void)
{
    if (m_pFPGADDAPI)
    {
        delete m_pFPGADDAPI;
        m_pFPGADDAPI = NULL;
    }
}

/// @fn         FPGADAPI::getDeviceList(list<string>  *out_pList)
///
/// @brief      Get the list of FPGA devices
/// @param      out_pFPGAList is a pointer to a list for FPGA devices
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADAPI::getDeviceList(list<string>  *out_pList)
{
    ErrorMsg*  pErrorMsg = NO_ERROR_MSG;

    if ((pErrorMsg = m_pFPGADDAPI->getDeviceList(out_pList)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_FPGA_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         FPGADAPI::getSPADInfo(string in_strDeviceName,
///                                   UINT32 *out_u32pStartOffset,
///                                   UINT32 *out_u32pTotalSize)
/// @brief      Get information about FPGA's scratch pad registers
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      out_u32pStartOffset is a pointer to a UINT32 for start register offset
/// @param      out_u32pTotalSize is a pointer to a UINT32 for size of registers
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADAPI::getSPADInfo(string in_strDeviceName,
                      UINT32 *out_u32pStartOffset,
                      UINT32 *out_u32pTotalSize)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;

    pErrorMsg = m_pFPGADDAPI->getSPADInfo(in_strDeviceName, out_u32pStartOffset, out_u32pTotalSize);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        returnErrorMsg(ERROR_FPGA_DAPI_GET_SPAD_INFO, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         FPGADAPI::readReg(string  in_strDeviceName,
///                               UINT32  in_u32Offset,
///                               string  *out_strpValue)
/// @brief      Read value from FPGA's register
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      in_u32Offset is a UINT32 value for register offset
/// @param      out_strpValue is a pointer to string for register value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADAPI::readReg(string  in_strDeviceName,
                  UINT32  in_u32Offset,
                  string  *out_strpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT8       u8Value;

    if (out_strpValue == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DAPI_READ_REG | ERROR_BAD_PARAMETER_THREE),
                              "NULL pointer");
    }

    if ((pErrorMsg = m_pFPGADDAPI->read(in_strDeviceName, in_u32Offset, &u8Value)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_FPGA_DAPI_READ_REG, NULL, pErrorMsg);
    }

    *out_strpValue = Int2HexStr((UINT32)u8Value, 2);

    return NO_ERROR_MSG;
}

/// @fn         FPGADAPI::readSPAD(string  in_strDeviceName,
///                               UINT32  in_u32Offset,
///                               string  *out_strpValue)
/// @brief      Read value from FPGA's scratch pad register
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      in_u32Offset is a UINT32 value for register offset
/// @param      out_strpValue is a pointer to string for register value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADAPI::readSPAD(string  in_strDeviceName,
                   UINT32  in_u32Offset,
                   string  *out_strpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT8       u8Value;

    if (out_strpValue == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DAPI_READ_SPAD | ERROR_BAD_PARAMETER_THREE),
                              "NULL pointer");
    }

    if ((pErrorMsg = m_pFPGADDAPI->readSPAD(in_strDeviceName, in_u32Offset, &u8Value)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_FPGA_DAPI_READ_SPAD, NULL, pErrorMsg);
    }

    *out_strpValue = Int2HexStr((UINT32)u8Value, 2);

    return NO_ERROR_MSG;
}

/// @fn         FPGADAPI::writeReg(string  in_strDeviceName,
///                                UINT32  in_u32Offset,
///                                string  in_strValue)
/// @brief      Write value to FPGA's register
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      in_u32Offset is a UINT32 value for register offset
/// @param      in_strValue is a string for register value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADAPI::writeReg(string  in_strDeviceName,
                   UINT32  in_u32Offset,
                   string  in_strValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT32      u32Value = std::stoul(in_strValue, nullptr, 0);

    if ((pErrorMsg = m_pFPGADDAPI->write(in_strDeviceName, in_u32Offset, u32Value)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_FPGA_DAPI_WRITE_REG, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         FPGADAPI::writeSPAD(string  in_strDeviceName,
///                                UINT32  in_u32Offset,
///                                string  in_strValue)
/// @brief      Write value to FPGA's scratch pad register
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      in_u32Offset is a UINT32 value for register offset
/// @param      in_strValue is a string for register value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADAPI::writeSPAD(string  in_strDeviceName,
                    UINT32  in_u32Offset,
                    string  in_strValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT8       u8Value = (UINT8)std::stoul(in_strValue, nullptr, 0);

    if ((pErrorMsg = m_pFPGADDAPI->writeSPAD(in_strDeviceName, in_u32Offset, u8Value)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_FPGA_DAPI_WRITE_SPAD, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}
