/// @file       MemoryDAPI.cpp
/// @brief      MemoryDAPI class
/// @details    Provide diagnostic and utility APIs for memory module
/// @version    1.0.0
/// @date       2014.01.29
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include <string>
#include <fstream>
#include "CommonFunctions.hpp"
#include "MemoryDAPI.hpp"
#include "SPDDDAPI.hpp"

/// @fn         MemoryDAPI::MemoryDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
/// @brief      MemoryDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
MemoryDAPI::MemoryDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
{
    if ((m_pMemoryDDAPI = new MemoryDDAPI("MEMORY_DRVR")) == NULL)
    {
        addErrorMsg( (ERRORCODE)(ERROR_MEMORY_DAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_API_CONSTRUCTOR1_ERROR), "Create MEMORY_DDAPI fail");
    }

    m_s32CPUAmount = 1;
    
    m_reservedMemoryList.clear();

    m_DIMMInfoList.clear();
}

/// @fn         MemoryDAPI::~MemoryDAPI(void)
/// @brief      MemoryDAPI destructor
/// @exception  Nothing
/// @return     Nothing
MemoryDAPI::~MemoryDAPI(void)
{
    m_DIMMInfoList.clear();
    
    m_reservedMemoryList.clear();
    
    // Save the error code
    if (m_pMemoryDDAPI != NULL)
    {
        delete m_pMemoryDDAPI;    
    }
}

/// @fn         MemoryDAPI::alignMemoryAddress(PHYSICAL_ADDRESS_T   in_physicalAddress,
///                                            INT32                in_s32ElementSize)
/// @brief      Align memory address for element size
/// @param      in_physicalAddress is a physical address
/// @param      in_s32ElementSize is an integer for element size
/// @execption  Nothing
/// @return     A physical memory address
PHYSICAL_ADDRESS_T
MemoryDAPI::alignMemoryAddress(PHYSICAL_ADDRESS_T   in_physicalAddress,
                               INT32                in_s32ElementSize)
{
    PHYSICAL_ADDRESS_T  alignmentBase = ((PHYSICAL_ADDRESS_T)in_s32ElementSize >> 3) - 1;

    if (in_physicalAddress & alignmentBase)
    {
        in_physicalAddress &= ~alignmentBase;
        in_physicalAddress += (PHYSICAL_ADDRESS_T)(in_s32ElementSize >> 3);
    }

    return in_physicalAddress;
}

/// @fn         MemoryDAPI::checkDeviceBySMBIOS(list<string>* inout_pDeviceList)
///
/// @brief      Check DIMM enable/disable status by SMBIOS
/// @param      inout_pDeviceList is a pointer to the list of strings
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::checkDeviceBySMBIOS(list<string>* inout_pDeviceList)
{
    CHAR    szTemp[256];
    FILE    *pFile;
    list<string> DMIInfoList;
    list<string>::iterator itDev;
    list<string>::iterator itInfo;
    INT32   s32Index, i;

    sprintf(szTemp, "dmidecode -t 17 | grep Size | awk '{print $2}'");
    if ((pFile = popen(szTemp, "r")) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_CHECK_DEVICE_BY_SMBIOS | ERROR_OPEN_FILE));
    }

    DMIInfoList.clear();

    while (fgets(szTemp, sizeof(szTemp), pFile) != NULL)
    {
        DMIInfoList.push_back(trimRightChar(szTemp, '\n'));
    }

    pclose(pFile);

    for (itDev = inout_pDeviceList->begin(); itDev != inout_pDeviceList->end(); itDev++)
    {
        s32Index = 0;
        // FIXME: This is a hard-coded well-known name for Aurora only
        sscanf(itDev->c_str(), "dimm%d", &s32Index);

        for (itInfo = DMIInfoList.begin(), i = 1; itInfo != DMIInfoList.end(); itInfo++, i++)
        {
            if (i == s32Index)
            {
                if ((*itInfo == "No") || (*itInfo == "0"))
                {
                    itDev->append("(disabled by BIOS)");
                }

                break;
            }
        }
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::getDeviceList(list<string>* out_pDeviceList
///                                       BOOL          in_bDetecedOnly)
/// @brief      Get the of DIMMs
/// @param      out_pDeviceList is a pointer to the list of strings
/// @param      in_bDetecedOnly is a bool value: TRUE - list detected devices FALSE - list supported devices
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::getDeviceList(list<string>* out_pDeviceList,
                          BOOL          in_bDetecedOnly)
{
    ErrorMsg*       pErrorMsg = NO_ERROR_MSG;
    SPDDDAPI*       pSPDDDAPI = NULL;
    list<string>::iterator itDev;

    if (out_pDeviceList == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_MEMORY_DAPI_GET_DEVICE_LIST | ERROR_BAD_PARAMETER),
                              "NULL pointer");
    }

    try
    {
        pSPDDDAPI = new SPDDDAPI();
        if ((pErrorMsg = pSPDDDAPI->getErrorMsg()))
        {
            addErrorMsg(ERROR_MEMORY_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
            delete pSPDDDAPI;
            return getErrorMsg();
        }

        if ((pErrorMsg = pSPDDDAPI->getDeviceList(out_pDeviceList, in_bDetecedOnly)))
        {
            addErrorMsg(ERROR_MEMORY_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
            delete pSPDDDAPI;
            return getErrorMsg();
        }

        // Get DIMM's device name based on SPD device name (SPD-XXX)
        for (itDev = out_pDeviceList->begin(); itDev != out_pDeviceList->end(); itDev++)
        {
            itDev->erase(0, 4);
        }

        // Do not have to check if the list is empty because SPD ddapi will
        // return error for this case

        delete pSPDDDAPI;

        if (in_bDetecedOnly)
        {
            if ((pErrorMsg = checkDeviceBySMBIOS(out_pDeviceList)))
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
            }
        }
    }
    catch (std::bad_alloc& e)
    {
        printf("Exception: %s\n", e.what());
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::getInformation(string           in_strDIMM,
///                                        DIMMInformation* out_pDIMMInfo)
/// @brief      Get memory information for the given device name
/// @param      in_strDIMM is a string for the device name
/// @param      out_pDIMMInfoList is a pointer to DIMMInformation
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::getInformation(string           in_strDIMM,
                           DIMMInformation* out_pDIMMInfo)
{
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    SPDDDAPI    spdDDAPI;
    DDRType     eDDRType = DDR_TYPE_UNKNOWN;
    UINT64      u64Temp;
    UINT32      u32Temp;

    if (out_pDIMMInfo == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_MEMORY_DAPI_GET_INFORMATION | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer");
    }

    if ((pErrorMsg = spdDDAPI.getErrorMsg()))
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    in_strDIMM = "spd-" + in_strDIMM;

    pErrorMsg = spdDDAPI.getDeviceType(in_strDIMM, &eDDRType);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    switch (eDDRType)
    {
        case DDR3:
            out_pDIMMInfo->m_strMemoryType = "DDR3";
            break;
        case DDR4:
            out_pDIMMInfo->m_strMemoryType = "DDR4";
            break;
        default:
            // This case has been handled
            break;
    }

    pErrorMsg = spdDDAPI.getModuleType(in_strDIMM, &out_pDIMMInfo->m_strModuleType);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    pErrorMsg = spdDDAPI.getSize(in_strDIMM, &u64Temp);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    u64Temp >>= 20;

    out_pDIMMInfo->m_strSize = to_string(u64Temp);
    out_pDIMMInfo->m_strSize += " MB";

    pErrorMsg = spdDDAPI.getRankCount(in_strDIMM, &u32Temp);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    out_pDIMMInfo->m_strRank = to_string(u32Temp);

    pErrorMsg = spdDDAPI.getSpeed(in_strDIMM, &u32Temp);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    out_pDIMMInfo->m_strMaxSpeed = to_string(u32Temp);
    out_pDIMMInfo->m_strMaxSpeed += " MHz";

    pErrorMsg = spdDDAPI.getModuleVendor(in_strDIMM, &out_pDIMMInfo->m_strModuleManufacturer);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    pErrorMsg = spdDDAPI.getDRAMVendor(in_strDIMM, &out_pDIMMInfo->m_strDRAMManufacturer);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    pErrorMsg = spdDDAPI.getPartNumber(in_strDIMM, &out_pDIMMInfo->m_strPartNumber);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    pErrorMsg = spdDDAPI.getSerialNumber(in_strDIMM, &out_pDIMMInfo->m_strSerialNumber);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_INFORMATION, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::getMemoryBuffer(UINT32              in_u32RequestedSize,
///                                         SYSTEM_ADDRESS_T*   out_pBufferAddress)
/// @brief      Get memory buffer for specified memory size 
/// @param      in_u32RequestedSize is a 32bits value for requested memory size
/// @param      out_pBufferAddress is a pointer to system memory address of allocated memory buffer
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::getMemoryBuffer(UINT32              in_u32RequestedSize,
                            SYSTEM_ADDRESS_T*   out_pBufferAddress)
{
    ErrorMsg* pErrorMsg = NULL;

    if (in_u32RequestedSize == 0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_GET_MEMORY_BUFFER | ERROR_BAD_PARAMETER_ONE),
                              "Memory size can not be zero",
                              NULL);
    }
    
    if (out_pBufferAddress == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_GET_MEMORY_BUFFER | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }
        
    pErrorMsg = m_pMemoryDDAPI->getMemoryBuffer(in_u32RequestedSize, out_pBufferAddress);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_MEMORY_BUFFER, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::getTestableMemory(list<MemoryRange>* out_pTestableMemoryList,
///                                           PHYSICAL_ADDRESS_T in_u64StartAddress,
///                                           PHYSICAL_ADDRESS_T in_u64EndAddress
/// @brief      Get testable memory information
/// @param      out_pReservedMemoryList is a pointer to string list for testable memory range
/// @param      in_u64StartAddress is a physical address for checked start address 
/// @param      in_u64EndAddress is a physical address for checked end address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::getTestableMemory(list<MemoryRange>* out_pTestableMemoryList,
                              PHYSICAL_ADDRESS_T in_u64StartAddress,
                              PHYSICAL_ADDRESS_T in_u64EndAddress)
{
    ErrorMsg*       pErrorMsg = NO_ERROR_MSG;
    FILE*           pFile;
    string          strError;
    CHAR            szLine[128], *chpTemp;
    INT32           i;
    list<INT32>     tempList;
    list< list <INT32> >::iterator  itList;
    list<MemoryRange>::iterator itRange;
    string          strTemp;

    if (out_pTestableMemoryList == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_MEMROY_DAPI_GET_TESTABLE_MEMORY | ERROR_BAD_PARAMETER_ONE),
                              "Testable memory list is a NULL",
                              NULL);
    }

    // Initialize testable memory ranges if there is no testable memory information
    strTemp = TEMP_FILE_TESTABLE_MEMORY;
    if ((pFile = fopen(strTemp.c_str(), "r")) == NULL)
    {
        if ((pErrorMsg = initializeTestableMemoryFile()) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_MEMROY_DAPI_GET_TESTABLE_MEMORY, NULL, pErrorMsg);
        }

        if ((pFile = fopen(strTemp.c_str(), "r")) == NULL)
        {
            strError = "can not open " + strTemp;
            return returnErrorMsg(ERRORCODE(ERROR_MEMROY_DAPI_GET_TESTABLE_MEMORY | ERROR_OPEN_FILE),
                                  strError.c_str(),
                                  NULL);
        }
    }

    out_pTestableMemoryList->clear();

    while (fgets(szLine, sizeof(szLine), pFile) != NULL)
    {
        if ((chpTemp = strstr(szLine, ":")) == NULL)
            continue;

        MemoryRange tempRange;

        tempRange.m_name = strtok(szLine, ":");

        sscanf(++chpTemp, "0x%llx-0x%llx", &tempRange.m_startAddress, &tempRange.m_endAddress);

        // Get socket list
        fgets(szLine, sizeof(szLine), pFile);

        chpTemp = strstr(szLine, ":");
        // There is one socket at least
        chpTemp = strtok(++chpTemp, ",");
        do
        {
            tempRange.m_controllerList.push_back(atoi(chpTemp));
        } while ((chpTemp = strtok(NULL, ",")) != NULL);

        // Get channel list for each socket
        for (i = 0; i < (INT32)tempRange.m_controllerList.size(); i++)
        {
            tempList.clear();

            // Get channel list
            fgets(szLine, sizeof(szLine), pFile);

            chpTemp = strstr(szLine, ":");
            // There is one channel at least
            chpTemp = strtok(++chpTemp, ",");
            do
            {
                tempList.push_back(atoi(chpTemp));
            } while ((chpTemp = strtok(NULL, ",")) != NULL);
            tempRange.m_channelList.push_back(tempList);
        }

        // Get rank list for each channel
        for (itList = tempRange.m_channelList.begin(); itList != tempRange.m_channelList.end(); itList++)
        {
            for (i = 0; i < (INT32)itList->size(); i++)
            {
                tempList.clear();
                // Get rank list
                fgets(szLine, sizeof(szLine), pFile);
                chpTemp = strstr(szLine, ":");
                // There is one rank at least
                chpTemp = strtok(++chpTemp, ",");
                do
                {
                    tempList.push_back(atoi(chpTemp));
                } while ((chpTemp = strtok(NULL, ",")) != NULL);
                tempRange.m_rankList.push_back(tempList);
            }
        }

        out_pTestableMemoryList->push_back(tempRange);
    }

    fclose(pFile);

    if ((in_u64StartAddress != 0) || (in_u64EndAddress != 0))
    {
        itRange = out_pTestableMemoryList->begin();
        while (itRange != out_pTestableMemoryList->end())
        {
            //        User
            // |---| |---| |---|
            //   X           X
            if ((itRange->m_endAddress < in_u64StartAddress) ||
                (in_u64EndAddress < itRange->m_startAddress))
            {
                itRange = out_pTestableMemoryList->erase(itRange);
            }
            //     User        User         User
            //    |----|      |----|       |----|
            // |----|    or  |-----|  or  |------|
            else if (itRange->m_startAddress < in_u64StartAddress)
            {
                itRange->m_startAddress = in_u64StartAddress;

                if (in_u64EndAddress < itRange->m_endAddress)
                {
                    itRange->m_endAddress = in_u64EndAddress;
                }

                itRange++;
            }
            //  User       User        User
            // |----|     |----|      |----|
            // |---|  or  |----|  or  |-----|
            //  |--|  or   |---|  or   |----|
            else
            {
                if (in_u64EndAddress < itRange->m_endAddress)
                {
                    itRange->m_endAddress = in_u64EndAddress;
                }

                itRange++;
            }
        }
    }
    
    if (out_pTestableMemoryList->empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_MEMROY_DAPI_GET_TESTABLE_MEMORY | ERROR_MEMORY_DAPI_NO_TESTABLE_MEMORY),
                                      NULL,
                                      NULL);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::getTestBlockSize(void)
/// @brief      Get test block size for multi-threading
/// @exception  Nothing
/// @return     An UINT32 value for block size
UINT32
MemoryDAPI::getTestBlockSize(void)
{
    UINT32 u32CPUCount = sysconf(_SC_NPROCESSORS_ONLN);
    UINT32 u32TargetSize = TOTAL_ALLOCATABLE_MEMORY_IN_MB / u32CPUCount;
    UINT32 u32CurrentSize = 1;

    while (u32CurrentSize < u32TargetSize)
    {
        u32CurrentSize <<= 1;
    }

    if (u32CurrentSize > u32TargetSize)
    {
        u32CurrentSize >>= 1;
    }

    return (u32CurrentSize << 20);
}

/// @fn         MemoryDAPI::readMemory8(PHYSICAL_ADDRESS_T in_physicalAddress)
/// @brief      Read 8bits data from specified physical address
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A 8bits data
UINT8
MemoryDAPI::readMemory8(PHYSICAL_ADDRESS_T in_physicalAddress)
{
    UINT8 u8Value = 0;

    u8Value = m_pMemoryDDAPI->read8(in_physicalAddress);
    if (m_pMemoryDDAPI->getErrorMsg() != NO_ERROR_MSG)
    {
        addErrorMsg(ERROR_MEMORY_DAPI_READ_MEMORY, NULL, m_pMemoryDDAPI->getErrorMsg());
    }

    return u8Value;
}

/// @fn         MemoryDAPI::readMemory16(PHYSICAL_ADDRESS_T in_physicalAddress)
/// @brief      Read 16bits data from specified physical address
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A 16bits data
UINT16
MemoryDAPI::readMemory16(PHYSICAL_ADDRESS_T in_physicalAddress)
{
    UINT16 u16Value = 0;

    u16Value = m_pMemoryDDAPI->read16(in_physicalAddress);
    if (m_pMemoryDDAPI->getErrorMsg() != NO_ERROR_MSG)
    {
        addErrorMsg(ERROR_MEMORY_DAPI_READ_MEMORY, NULL, m_pMemoryDDAPI->getErrorMsg());
    }

    return u16Value;
}

/// @fn         MemoryDAPI::readMemory32(PHYSICAL_ADDRESS_T in_physicalAddress)
/// @brief      Read 32bits data from specified physical address
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A 32bits data
UINT32
MemoryDAPI::readMemory32(PHYSICAL_ADDRESS_T in_physicalAddress)
{
    UINT32 u32Value = 0;

    u32Value = m_pMemoryDDAPI->read32(in_physicalAddress);
    if (m_pMemoryDDAPI->getErrorMsg() != NO_ERROR_MSG)
    {
        addErrorMsg(ERROR_MEMORY_DAPI_READ_MEMORY, NULL, m_pMemoryDDAPI->getErrorMsg());
    }

    return u32Value;
}

/// @fn         MemoryDAPI::readMemory64(PHYSICAL_ADDRESS_T in_physicalAddress)
/// @brief      Read 64bits data from specified physical address
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A 64bits data
UINT64
MemoryDAPI::readMemory64(PHYSICAL_ADDRESS_T in_physicalAddress)
{
    UINT64 u64Value = 0;

    u64Value = m_pMemoryDDAPI->read64(in_physicalAddress);
    if (m_pMemoryDDAPI->getErrorMsg() != NO_ERROR_MSG)
    {
        addErrorMsg(ERROR_MEMORY_DAPI_READ_MEMORY, NULL, m_pMemoryDDAPI->getErrorMsg());
    }

    return u64Value;
}

/// @fn         MemoryDAPI::releaseMemoryBuffer(SYSTEM_ADDRESS_T in_bufferAddress)
/// @brief      Release memory buffer for specified memory buffer address
/// @param      in_bufferAddress is a memory address for specified memory buffer
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::releaseMemoryBuffer(SYSTEM_ADDRESS_T in_bufferAddress)
{
    ErrorMsg* pErrorMsg = NULL;
    
    if ((SYSTEM_ADDRESS_T *)in_bufferAddress == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_RELEASE_MEMORY_BUFFER | ERROR_BAD_PARAMETER),
                              "Invalid buffer address",
                              NULL);
    }

    pErrorMsg = m_pMemoryDDAPI->releaseMemoryBuffer(in_bufferAddress);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_RELEASE_MEMORY_BUFFER, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::writeMemory8(PHYSICAL_ADDRESS_T    in_physicalAddress,
///                                      UINT8                 in_u8Value)
/// @brief      Write 8bits data to specified physical address
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::writeMemory8(PHYSICAL_ADDRESS_T    in_physicalAddress,
                         UINT8                 in_u8Value)
{
    ErrorMsg* pErrorMsg = NULL;

    pErrorMsg = m_pMemoryDDAPI->write(in_physicalAddress, in_u8Value);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_WRITE_MEMORY, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @overload   MemoryDAPI::writeMemory16(PHYSICAL_ADDRESS_T    in_physicalAddress,
///                                       UINT16                in_u16Value)
/// @brief      Write 16bits data to specified physical address
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::writeMemory16(PHYSICAL_ADDRESS_T    in_physicalAddress,
                          UINT16                in_u16Value)
{
    ErrorMsg* pErrorMsg = NULL;

    pErrorMsg = m_pMemoryDDAPI->write(in_physicalAddress, in_u16Value);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_WRITE_MEMORY, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @overload   MemoryDAPI::writeMemory32(PHYSICAL_ADDRESS_T    in_physicalAddress,
///                                       UINT32                in_u32Value)
/// @brief      Write 32bits data to specified physical address
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::writeMemory32(PHYSICAL_ADDRESS_T    in_physicalAddress,
                          UINT32                in_u32Value)
{
    ErrorMsg* pErrorMsg = NULL;

    pErrorMsg = m_pMemoryDDAPI->write(in_physicalAddress, in_u32Value);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_WRITE_MEMORY, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @overload   MemoryDAPI::writeMemory64(PHYSICAL_ADDRESS_T    in_physicalAddress,
///                                       UINT64                in_u64Value)
/// @brief      Write 64bits data to specified physical address
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::writeMemory64(PHYSICAL_ADDRESS_T    in_physicalAddress,
                          UINT64                in_u64Value)
{
    ErrorMsg* pErrorMsg = NULL;

    pErrorMsg = m_pMemoryDDAPI->write(in_physicalAddress, in_u64Value);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_WRITE_MEMORY, NULL, pErrorMsg);
    }
    
    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::getOverlapMemoryRange(MemoryRange in_memoryRangeA,
///                                               MemoryRange in_memoryRangeB)
/// @brief      Get overlap memory range of two memory ranges
/// @param      in_memoryRangeA is a memory range
/// @param      in_memoryRangeB is a memory range
/// @exception  Nothing
/// @return     A MemoryRange for overlap memory range
MemoryRange
MemoryDAPI::getOverlapMemoryRange(MemoryRange in_memoryRangeA,
                                  MemoryRange in_memoryRangeB)
{
    MemoryRange overlapMemoryRange;

    if (in_memoryRangeA.m_startAddress <= in_memoryRangeB.m_startAddress)
    {
        if (in_memoryRangeA.m_endAddress > in_memoryRangeB.m_startAddress)
        {
            if (in_memoryRangeA.m_endAddress <= in_memoryRangeB.m_endAddress)
            {
                //      -- A start address
                //       |
                //      -- B start address
                //       |
                //       |  (Overlapping)
                //       |
                //      -- A end address
                //       |
                //      -- B end address
                overlapMemoryRange.m_startAddress = in_memoryRangeB.m_startAddress;
                overlapMemoryRange.m_endAddress = in_memoryRangeA.m_endAddress;
            }
            else
            {
                //      -- A start address
                //       |
                //      -- B start address
                //       |
                //       |  (Overlapping)
                //       |
                //      -- B end address
                //       |
                //      -- A end address
                overlapMemoryRange.m_startAddress = in_memoryRangeB.m_startAddress;
                overlapMemoryRange.m_endAddress = in_memoryRangeB.m_endAddress;
            }
        }
    }
    else
    {
        if (in_memoryRangeB.m_endAddress > in_memoryRangeA.m_startAddress)
        {
            if (in_memoryRangeB.m_endAddress <= in_memoryRangeA.m_endAddress)
            {
                //      -- B start address
                //       |
                //      -- A start address
                //       |
                //       |  (Overlapping)
                //       |
                //      -- B end address
                //       |
                //      -- A end address
                overlapMemoryRange.m_startAddress = in_memoryRangeA.m_startAddress;
                overlapMemoryRange.m_endAddress = in_memoryRangeB.m_endAddress;
            }
            else
            {
                //      -- B start address
                //       |
                //      -- A start address
                //       |
                //       |  (Overlapping)
                //       |
                //      -- A end address
                //       |
                //      -- B end address
                overlapMemoryRange.m_startAddress = in_memoryRangeA.m_startAddress;
                overlapMemoryRange.m_endAddress = in_memoryRangeA.m_endAddress;
            }
        }
    }

    return overlapMemoryRange;
}

/// @fn         MemoryDAPI::getRankRange(list<MemoryRange>* out_pRangeList)
///
/// @brief      Get memory ranges according to ranks of DIMMs
/// @param      out_pRangeList is a pointer to the list of MemoryRange
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::getRankRange(list<MemoryRange>* out_pRangeList)
{
    ErrorMsg*       pErrorMsg = NO_ERROR_MSG;
    SPDDDAPI        spdDDAPI;
    list<string>    spdList;
    string          strDIMM;
    MemoryRange     dimmRange, tempRange;
    UINT64          u64Size;
    UINT32          u32RankCount;
    INT32           s32Rank;

    // Do not have to check if out_pRangeList is a NULL pointer because
    // this is a private function

    // Check if errors occur in SPD DDAPI constructor
    if ((pErrorMsg = spdDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_RANK_RANGE, NULL, pErrorMsg);
    }

    spdList.clear();

    // Get SPD list
    pErrorMsg = spdDDAPI.getDeviceList(&spdList);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_GET_RANK_RANGE, NULL, pErrorMsg);
    }

    dimmRange.m_startAddress = 0ULL;
    dimmRange.m_endAddress = -1ULL;

    // Initialize memory range list
    while (!spdList.empty())
    {
        strDIMM = spdList.front();

        pErrorMsg = spdDDAPI.getSize(strDIMM, &u64Size);
        if (pErrorMsg != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_MEMORY_DAPI_GET_RANK_RANGE, NULL, pErrorMsg);
        }

        pErrorMsg = spdDDAPI.getRankCount(strDIMM, &u32RankCount);
        if (pErrorMsg != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_MEMORY_DAPI_GET_RANK_RANGE, NULL, pErrorMsg);
        }

        dimmRange.m_startAddress = dimmRange.m_endAddress + 1;
        dimmRange.m_endAddress = dimmRange.m_startAddress + (PHYSICAL_ADDRESS_T)u64Size - 1;

        s32Rank = 0;
        u64Size = u64Size / (UINT64)u32RankCount;

        // Handle range for each rank
        tempRange.m_name = strDIMM;
        while (s32Rank < (INT32)u32RankCount)
        {
            tempRange.m_startAddress = dimmRange.m_startAddress + (PHYSICAL_ADDRESS_T)(u64Size * (UINT64)s32Rank);
            tempRange.m_endAddress = tempRange.m_startAddress + (PHYSICAL_ADDRESS_T)u64Size - 1;

            out_pRangeList->push_back(tempRange);

            s32Rank++;
        }

        spdList.pop_front();
    }

    // Do not have to check if the list of range is empty because error cases
    // have been handled in above steps

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::initializeMemoryLayoutFile(void)
///
/// @brief      Initialize the file for memory layouts
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::initializeMemoryLayoutFile(void)
{
    ErrorMsg*       pErrorMsg = NO_ERROR_MSG;
    MemoryRange     highRamRange, remapRange, tempRange, reservedRange;
    list<MemoryRange> rangeList, remapRangeList;
    list<MemoryRange>::iterator itRange;
    string          strDIMM, strTemp;
    PHYSICAL_ADDRESS_T tempAddress = 0, remapSize = 0;
    FILE*           pFile = NULL;

    // Initialize memory range list
    pErrorMsg = getRankRange(&rangeList);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_INITIALIZE_MEMORY_LAYOUT_FILE, NULL , pErrorMsg);
    }

    switch (getArch())
    {
        case ARCH_X86:
            pErrorMsg = m_pMemoryDDAPI->getX86HighRamRange(&highRamRange);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_INITIALIZE_MEMORY_LAYOUT_FILE,
                                      NULL,
                                      pErrorMsg);
            }

            strDIMM.clear();

            // Find out ranges which should be remapped
            itRange = rangeList.begin();
            while (itRange != rangeList.end())
            {
                // Save start address of one DIMM
                if (strDIMM != itRange->m_name)
                {
                    strDIMM = itRange->m_name;
                    tempAddress = itRange->m_startAddress;
                }

                // ------
                // |    |
                // |DIMM|   -----------
                // |    |   HIGH_MEMORY
                if (tempAddress < highRamRange.m_startAddress)
                {
                    // -------
                    // |RANGE|
                    // -------   HIGH_MEMORY
                    //           -----------
                    if (itRange->m_endAddress < highRamRange.m_startAddress)
                    {
                        itRange++;
                        continue;
                    }

                    // Handle ranges which should be remapped
                    // -------
                    // |     |
                    // |RANGE|   ----------
                    // |     |   HIGH_MEMORY
                    // -------
                    if (itRange->m_startAddress < highRamRange.m_startAddress)
                    {
                        remapRange.m_name = itRange->m_name;
                        remapRange.m_startAddress = 0ULL;
                        remapRange.m_endAddress = itRange->m_endAddress - highRamRange.m_startAddress;
                        remapRangeList.push_back(remapRange);

                        remapSize += remapRange.getSize();

                        itRange->m_endAddress = highRamRange.m_startAddress - 1;
                        itRange++;
                    }
                    // -------   ------ HIGH_MEMORY
                    // |     |
                    // |RANGE|
                    // |     |
                    // -------
                    else
                    {
                        remapRange.m_name = itRange->m_name;
                        remapRange.m_startAddress = 0ULL;
                        remapRange.m_endAddress = itRange->getSize() - 1;
                        remapRangeList.push_back(remapRange);

                        remapSize += remapRange.getSize();

                        itRange = rangeList.erase(itRange);
                    }
                }
                // ------   -----------
                // |    |   HIGH_MEMORY
                // |DIMM|
                // |    |
                // ------
                else
                {
                    // Shift start address and end address for remapped memory range
                    itRange->m_startAddress -= remapSize;
                    itRange->m_endAddress -= remapSize;

                    itRange++;
                }
            } // End of find out remapped ranges

            // No range for DIMMs
            if (rangeList.empty())
            {
                return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_INITIALIZE_MEMORY_LAYOUT_FILE | ERROR_MEMORY_DAPI_NO_DIMM_INFO));
            }

            // Handle remapped ranges which will be remapped behind the last DIMM
            itRange = remapRangeList.begin();
            while (itRange != remapRangeList.end())
            {
                itRange->m_startAddress += rangeList.back().m_endAddress + 1;
                itRange->m_endAddress += rangeList.back().m_endAddress + 1;

                rangeList.push_back(*itRange);
                itRange++;
            }

            // Calculate x86 MMIO reserved memory
            if (highRamRange.m_endAddress > rangeList.back().m_endAddress)
            {
                remapRange.m_startAddress = rangeList.back().m_endAddress + 1;
                remapRange.m_endAddress = highRamRange.m_endAddress;

                reservedRange.m_endAddress = highRamRange.m_startAddress - 1;
                reservedRange.m_startAddress = (reservedRange.m_endAddress - remapRange.getSize()) + 1;

                // Check each range if it is in the MMIO reserved range
                itRange = rangeList.begin();
                while (itRange != rangeList.end())
                {
                    // -------
                    // |RANGE|
                    // -------   ----------
                    //           |Reserved|
                    if (itRange->m_endAddress < reservedRange.m_startAddress)
                    {
                        itRange++;
                        continue;
                    }

                    tempRange.m_name = itRange->m_name;
                    tempRange.m_startAddress = remapRange.m_startAddress;
                    remapRange.m_startAddress += (itRange->m_endAddress - reservedRange.m_startAddress) + 1;
                    tempRange.m_endAddress = remapRange.m_startAddress - 1;
                    rangeList.push_back(tempRange);

                    // -------
                    // |RANGE|   ----------
                    // |     |   |Reserved|
                    //
                    if (itRange->m_startAddress < reservedRange.m_startAddress)
                    {
                        // |RANGE|    |        |
                        // -------    |Reserved|
                        //            ----------
                        if (itRange->m_endAddress < reservedRange.m_endAddress)
                        {
                            tempAddress = itRange->m_endAddress;
                            itRange->m_endAddress = reservedRange.m_startAddress - 1;
                            reservedRange.m_startAddress = tempAddress + 1;

                            itRange++;
                        }
                        // |RANGE|    |Reserved|
                        // -------    ----------
                        else
                        {
                            itRange->m_endAddress = reservedRange.m_startAddress - 1;
                            break;
                        }
                    }
                    // -------   ----------
                    // |RANGE|   |Reserved|
                    else
                    {
                        // |RANGE|    |        |
                        // -------    |Reserved|
                        //            ----------
                        if (itRange->m_endAddress < reservedRange.m_endAddress)
                        {
                            reservedRange.m_startAddress = itRange->m_endAddress + 1;
                            itRange = rangeList.erase(itRange);
                        }
                        // |RANGE|    |Reserved|
                        // -------    ----------
                        else
                        {
                            itRange = rangeList.erase(itRange);
                            break;
                        }
                    }
                }
            } // Handle MMIO reserved ranges

            strTemp = "mkdir -p ";
            strTemp += MEMORY_TEMP_FILE_DIR;
            system(strTemp.c_str());

            //strTemp = MEMORY_TEMP_FILE_DIR;
            strTemp = TEMP_FILE_MEMORY_LAYOUT;
            if ((pFile = fopen(strTemp.c_str(), "w")) == NULL)
            {
                strTemp = "can not open " + strTemp;
                return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_INITIALIZE_MEMORY_LAYOUT_FILE | ERROR_OPEN_FILE),
                                      strTemp.c_str());
            }

            itRange = rangeList.begin();
            while (itRange != rangeList.end())
            {
                strTemp = itRange->m_name + ":";
                strTemp += UINT64ToHexStr(itRange->m_startAddress) + "," + UINT64ToHexStr(itRange->m_endAddress) + "\n";
                fwrite(strTemp.c_str(), sizeof(char), strTemp.length(), pFile);
                itRange = rangeList.erase(itRange);
            }

            fclose(pFile);

            break; // Handle for x86
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_INITIALIZE_MEMORY_LAYOUT_FILE | ERROR_UNKNOWN_ARCH),
                                  NULL,
                                  NULL);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::initializeTestableMemoryFile(void)
///
/// @brief      Initialize the file for testable memory ranges
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::initializeTestableMemoryFile(void)
{
    ErrorMsg            *pErrorMsg = NO_ERROR_MSG;
    FILE                *pInputFile = NULL, *pOutputFile = NULL;
    CHAR                szLine[128], chTemp, *chpTemp;
    MemoryRange         reservedRange, tempRange;
    string              strError, strTemp;
    BOOL                bHaveTestableRange = FALSE;
    INT32               i, s32Controllers, s32Channels;

    reservedRange.m_startAddress = reservedRange.m_endAddress = 0ULL;

    // Get reserved memory range for system
    if ((pInputFile = fopen("/proc/cmdline", "r")) == NULL)
        return returnErrorMsg((ERRORCODE)(ERROR_MEMROY_DAPI_GET_TESTABLE_MEMORY | ERROR_OPEN_FILE),
                               "can not open /proc/cmdline",
                               NULL);

    while (fgets(szLine, sizeof(szLine), pInputFile) != NULL)
    {
        if ((chpTemp = strstr(szLine, "mem=")) != NULL)
        {
            chpTemp += strlen("mem=");

            reservedRange.m_endAddress = atoll(chpTemp);

            while (isdigit(*(++chpTemp))) ;

            chTemp = *chpTemp;

            if ((chTemp == 'g') || (chTemp == 'G'))
            {
                reservedRange.m_endAddress <<= 30;
            }
            else if ((chTemp == 'm') || (chTemp == 'M'))
            {
                reservedRange.m_endAddress <<= 20;
            }
            else // k or K
            {
                reservedRange.m_endAddress <<= 10;
            }

            break;
        }
    }

    fclose(pInputFile);

    if (reservedRange.m_endAddress == (PHYSICAL_ADDRESS_T)0)
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_INITIALIZE_TESTABLE_MEMORY_FILE | ERROR_MEMORY_DAPI_NO_TESTABLE_MEMORY),
                              NULL,
                              NULL);

    reservedRange.m_endAddress--;

    strTemp = TEMP_FILE_MEMORY_LAYOUT;
    if ((pInputFile = fopen(strTemp.c_str(), "r")) == NULL)
    {
        if ((pErrorMsg = m_pMemoryDDAPI->initializeMemoryLayoutFile()) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_MEMORY_DAPI_INITIALIZE_TESTABLE_MEMORY_FILE, NULL, pErrorMsg);
        }

        if ((pInputFile = fopen(strTemp.c_str(), "r")) == NULL)
        {
            strError = "can not open " + strTemp;
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_INITIALIZE_TESTABLE_MEMORY_FILE | ERROR_OPEN_FILE),
                                  strError.c_str(),
                                  NULL);
        }
    }

    strTemp = TEMP_FILE_TESTABLE_MEMORY;
    if ((pOutputFile = fopen(strTemp.c_str(), "w")) == NULL)
    {
        fclose(pInputFile);
        strError = "can not open " + strTemp;
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_INITIALIZE_TESTABLE_MEMORY_FILE | ERROR_OPEN_FILE),
                              strError.c_str(),
                              NULL);
    }

    // Get memory range from memory layout file
    while (fgets(szLine, sizeof(szLine), pInputFile) != NULL)
    {
        // This is not a range
        if (strstr(szLine, "-") == NULL)
            continue;

        chpTemp = strstr(szLine, ":");
        sscanf(++chpTemp, "0x%llx-0x%llx", &tempRange.m_startAddress, &tempRange.m_endAddress);

        // -------
        // |RANGE|
        // -------
        //         ------ Reserved
        if (tempRange.m_endAddress <= reservedRange.m_endAddress)
            continue;

        bHaveTestableRange = TRUE;

        // -------
        // |RANGE|  ------ Reserved
        // -------
        if (tempRange.m_startAddress < reservedRange.m_endAddress)
        {
            tempRange.m_startAddress = reservedRange.m_endAddress + 1;

            strTemp = strtok(szLine, ":");
            strTemp += ":";
            strTemp += UINT64ToHexStr(tempRange.m_startAddress) + "-";
            strTemp += UINT64ToHexStr(tempRange.m_endAddress) + "\n";
            fwrite(strTemp.c_str(), sizeof(char), strTemp.length(), pOutputFile);
        }
        else
        {
            fwrite(szLine, sizeof(char), strlen(szLine), pOutputFile);
        }

        // There is one line for controller list at least
        fgets(szLine, sizeof(szLine), pInputFile);
        fwrite(szLine, sizeof(char), strlen(szLine), pOutputFile);

        // One controller at least
        s32Controllers = 1;
        chpTemp = strstr(szLine, ",");
        while (chpTemp)
        {
            s32Controllers++;
            chpTemp = strstr(++chpTemp, ",");
        }

        s32Channels = 0;
        for (i = 0; i < s32Controllers; i++)
        {
            fgets(szLine, sizeof(szLine), pInputFile);
            fwrite(szLine, sizeof(char), strlen(szLine), pOutputFile);

            // One channel at least
            s32Channels++;
            chpTemp = strstr(szLine, ",");
            while (chpTemp)
            {
                s32Channels++;
                chpTemp = strstr(++chpTemp, ",");
            }
        }

        for (i = 0; i < s32Channels; i++)
        {
            fgets(szLine, sizeof(szLine), pInputFile);
            fwrite(szLine, sizeof(char), strlen(szLine), pOutputFile);
        }
    }

    fclose(pOutputFile);
    fclose(pInputFile);

    if (!bHaveTestableRange)
    {
        // Remove testable memory file
        strError = "rm ";
        strError += TEMP_FILE_TESTABLE_MEMORY;
        system(strError.c_str());

        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_INITIALIZE_TESTABLE_MEMORY_FILE | ERROR_MEMORY_DAPI_NO_TESTABLE_MEMORY),
                              NULL,
                              NULL);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::verifyMemoryBlock(PHYSICAL_ADDRESS_T    in_startAddress,
///                                           PHYSICAL_ADDRESS_T    in_endAddress,
///                                           UINT8                 in_u8UserTestPattern)
/// @brief      Verify memory block for specified test pattern   
/// @param      in_startAddress is a physical address for start address
/// @param      in_endAddress is a physical address for end address
/// @param      in_u8UserTestPattern is a 8bits test pattern
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::verifyMemoryBlock(PHYSICAL_ADDRESS_T    in_startAddress,
                              PHYSICAL_ADDRESS_T    in_endAddress,
                              UINT8                 in_u8UserTestPattern)
{
    PHYSICAL_ADDRESS_T  currentAddress = in_startAddress;
    ErrorMsg*           pErrorMsg = NULL;
    INT32               s32Step = sizeof(UINT8);
    string              strErrorMsg;

    if (in_startAddress < in_endAddress)
    {
        // Write test pattern
        while (currentAddress < in_endAddress)
        {
            pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u8UserTestPattern);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
            }
            currentAddress += (PHYSICAL_ADDRESS_T)s32Step;
        }

        currentAddress = in_startAddress;
        // Read and verify test pattern
        while (currentAddress < in_endAddress)
        {
            if (in_u8UserTestPattern != m_pMemoryDDAPI->read8(currentAddress))
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           strErrorMsg.c_str(),
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
                }                
            }

            currentAddress += (PHYSICAL_ADDRESS_T)s32Step;
        }
    }
    else
    {
        // Write test pattern
        while (currentAddress > in_endAddress)
        {
            pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u8UserTestPattern);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
            }            
            currentAddress -= (PHYSICAL_ADDRESS_T)s32Step;
        } 

        currentAddress = in_startAddress;
        // Read and verify test pattern
        while (currentAddress < in_endAddress)
        {
            if (in_u8UserTestPattern != m_pMemoryDDAPI->read8(currentAddress))
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           strErrorMsg.c_str(),
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
                }                
            }

            currentAddress -= (PHYSICAL_ADDRESS_T)s32Step;
        }
    }

    return NO_ERROR_MSG;
}

/// @overload   MemoryDAPI::verifyMemoryBlock(PHYSICAL_ADDRESS_T    in_startAddress,
///                                           PHYSICAL_ADDRESS_T    in_endAddress,
///                                           UINT16                in_u16UserTestPattern)
/// @brief      Verify memory block for specified test pattern   
/// @param      in_startAddress is a physical address for start address
/// @param      in_endAddress is a physical address for end address
/// @param      in_u16UserTestPattern is a 16bits test pattern
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::verifyMemoryBlock(PHYSICAL_ADDRESS_T    in_startAddress,
                              PHYSICAL_ADDRESS_T    in_endAddress,
                              UINT16                in_u16UserTestPattern)
{
    PHYSICAL_ADDRESS_T  currentAddress = in_startAddress;
    ErrorMsg*           pErrorMsg = NULL;
    INT32               s32Step = sizeof(UINT16);
    string              strErrorMsg;    

    if (in_startAddress < in_endAddress)
    {
        // Write test pattern
        while (currentAddress < in_endAddress)
        {
            pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u16UserTestPattern);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
            }
            currentAddress += (PHYSICAL_ADDRESS_T)s32Step;
        }

        currentAddress = in_startAddress;
        // Read and verify test pattern
        while (currentAddress < in_endAddress)
        {
            if (in_u16UserTestPattern != m_pMemoryDDAPI->read16(currentAddress))
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           strErrorMsg.c_str(),
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
                }                
            }

            currentAddress += (PHYSICAL_ADDRESS_T)s32Step;
        }
    }
    else
    {
        // Write test pattern
        while (currentAddress > in_endAddress)
        {
            pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u16UserTestPattern);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
            }            
            currentAddress -= (PHYSICAL_ADDRESS_T)s32Step;
        } 

        currentAddress = in_startAddress;
        // Read and verify test pattern
        while (currentAddress < in_endAddress)
        {
            if (in_u16UserTestPattern != m_pMemoryDDAPI->read16(currentAddress))
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           strErrorMsg.c_str(),
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
                }                
            }

            currentAddress -= (PHYSICAL_ADDRESS_T)s32Step;
        }
    }

    return NO_ERROR_MSG;
}

/// @overload   MemoryDAPI::verifyMemoryBlock(PHYSICAL_ADDRESS_T    in_startAddress,
///                                           PHYSICAL_ADDRESS_T    in_endAddress,
///                                           UINT32                in_u32UserTestPattern)
/// @brief      Verify memory block for specified test pattern   
/// @param      in_startAddress is a physical address for start address
/// @param      in_endAddress is a physical address for end address
/// @param      in_u32UserTestPattern is a 32bits test pattern
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::verifyMemoryBlock(PHYSICAL_ADDRESS_T    in_startAddress,
                              PHYSICAL_ADDRESS_T    in_endAddress,
                              UINT32                in_u32UserTestPattern)
{
    PHYSICAL_ADDRESS_T  currentAddress = in_startAddress;
    ErrorMsg*           pErrorMsg = NULL;
    INT32               s32Step = sizeof(UINT32);
    string              strErrorMsg;    

    if (in_startAddress < in_endAddress)
    {
        // Write test pattern
        while (currentAddress < in_endAddress)
        {
            pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u32UserTestPattern);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
            }
            currentAddress += (PHYSICAL_ADDRESS_T)s32Step;
        }

        currentAddress = in_startAddress;
        // Read and verify test pattern
        while (currentAddress < in_endAddress)
        {
            if (in_u32UserTestPattern != m_pMemoryDDAPI->read32(currentAddress))
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           strErrorMsg.c_str(),
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
                }
            }

            currentAddress += (PHYSICAL_ADDRESS_T)s32Step;
        }
    }
    else
    {
        // Write test pattern
        while (currentAddress > in_endAddress)
        {
            pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u32UserTestPattern);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
            }            
            currentAddress -= s32Step;
        } 

        currentAddress = in_startAddress;
        // Read and verify test pattern
        while (currentAddress < in_endAddress)
        {
            if (in_u32UserTestPattern != m_pMemoryDDAPI->read32(currentAddress))
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           strErrorMsg.c_str(),
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
                }                
            }

            currentAddress -= s32Step;
        }
    }

    return NO_ERROR_MSG;
}

/// @overload   MemoryDAPI::verifyMemoryBlock(PHYSICAL_ADDRESS_T    in_startAddress,
///                                           PHYSICAL_ADDRESS_T    in_endAddress,
///                                           UINT64                in_u64UserTestPattern)
/// @brief      Verify memory block for specified test pattern   
/// @param      in_startAddress is a physical address for start address
/// @param      in_endAddress is a physical address for end address
/// @param      in_u64UserTestPattern is a 64bits test pattern
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::verifyMemoryBlock(PHYSICAL_ADDRESS_T    in_startAddress,
                              PHYSICAL_ADDRESS_T    in_endAddress,
                              UINT64                in_u64UserTestPattern)
{
    PHYSICAL_ADDRESS_T  currentAddress = in_startAddress;
    ErrorMsg*           pErrorMsg = NULL;
    INT32               s32Step = sizeof(UINT64);
    string              strErrorMsg;    

    if (in_startAddress < in_endAddress)
    {
        // Write test pattern
        while (currentAddress < in_endAddress)
        {
            pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u64UserTestPattern);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
            }
            currentAddress += (PHYSICAL_ADDRESS_T)s32Step;
        }

        currentAddress = in_startAddress;
        // Read and verify test pattern
        while (currentAddress < in_endAddress)
        {
            if (in_u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress))
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)                    
                {
                    strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           strErrorMsg.c_str(),
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
                }                
            }

            currentAddress += (PHYSICAL_ADDRESS_T)s32Step;
        }
    }
    else
    {
        // Write test pattern
        while (currentAddress > in_endAddress)
        {
            pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u64UserTestPattern);
            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
            }            
            currentAddress -= (PHYSICAL_ADDRESS_T)s32Step;
        } 

        currentAddress = in_startAddress;
        // Read and verify test pattern
        while (currentAddress < in_endAddress)
        {
            if (in_u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress))
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           strErrorMsg.c_str(),
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK, NULL, pErrorMsg);
                }                
            }

            currentAddress -= (PHYSICAL_ADDRESS_T)s32Step;
        }
    }

    return NO_ERROR_MSG;
}
