/// @file       MemoryDAPI_DataLineTest.cpp
/// @brief      Provide data line test for MemoryDAPI
/// @details    Verify memory data lines by write and read test pattern
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "MemoryDAPI.hpp"

/// @fn         MemoryDAPI::dataLineTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                      PHYSICAL_ADDRESS_T in_endAddress,
///                                      INT32              in_s32DataLineAmount)
/// @brief      Verify memory data lines
/// @param      in_startAddress is a physcial address for start address
/// @param      in_endAddress is a physcial address for end address
/// @param      in_s32DataLineAmount is 32bits value for the amount of data lines (ex: 8/16/32/64)
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::dataLineTest(PHYSICAL_ADDRESS_T in_startAddress,
                         PHYSICAL_ADDRESS_T in_endAddress,
                         INT32              in_s32DataLineAmount)
{
    ErrorMsg*           pErrorMsg = NULL;
    PHYSICAL_ADDRESS_T  testBlockSize = (PHYSICAL_ADDRESS_T)m_pMemoryDDAPI->getMemoryMapSize();
    PHYSICAL_ADDRESS_T  memoryBlockSize = 0;
    PHYSICAL_ADDRESS_T  verifiedMemorySize = 0;
    PHYSICAL_ADDRESS_T  startAddress = in_startAddress;
    PHYSICAL_ADDRESS_T  endAddress = in_endAddress;
    PHYSICAL_ADDRESS_T  currentAddress = in_startAddress;
    UINT64              u64UserTestPattern = 0; 

    switch (in_s32DataLineAmount)
    {
        case 8:
        case 16:
        case 32:
        case 64:
            // Round up memory address for alignment
            startAddress = alignMemoryAddress(startAddress, in_s32DataLineAmount);

            memoryBlockSize = endAddress - startAddress;
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_DATA_LINE_TEST | ERROR_BAD_PARAMETER_THREE), 
                                  "invalid data line amount",
                                  NULL);
            break;

    }
    
    if (memoryBlockSize == 0)
    {
        return NO_ERROR_MSG;
    }

    initProgress(memoryBlockSize);
    
    while (verifiedMemorySize < memoryBlockSize)
    {
      
        // The last test memory block size may be smaller than memory map size 
        if ((memoryBlockSize - verifiedMemorySize) < testBlockSize)
        {
            testBlockSize = memoryBlockSize - verifiedMemorySize;
        }
        
        // Write, read and verify test pattern for test block size
        switch (in_s32DataLineAmount)
        {
            case 8:
                pErrorMsg = verifyMemoryBlock(currentAddress, 
                                              (currentAddress + testBlockSize),
                                              (UINT8)u64UserTestPattern);                 
                break;
            case 16:
                pErrorMsg = verifyMemoryBlock(currentAddress, 
                                              (currentAddress + testBlockSize),
                                              (UINT16)u64UserTestPattern);                
                break;
            case 32:
                pErrorMsg = verifyMemoryBlock(currentAddress, 
                                              (currentAddress + testBlockSize),
                                              (UINT32)u64UserTestPattern);                
                break;
            case 64:
                pErrorMsg = verifyMemoryBlock(currentAddress, 
                                              (currentAddress + testBlockSize),
                                              (UINT64)u64UserTestPattern);
                break;
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
        }

        u64UserTestPattern = 1;

        switch (in_s32DataLineAmount)
        {
            case 8:
                while ((UINT8)u64UserTestPattern)
                {
                    pErrorMsg = verifyMemoryBlock(currentAddress, 
                                                  (currentAddress + testBlockSize),
                                                  (UINT8)u64UserTestPattern);                 

                    if (pErrorMsg != NO_ERROR_MSG)
                    {
                        initProgress(0);
                        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
                    }
                    else
                    {
                        u64UserTestPattern <<= 1;
                    }
                }
                break;
            case 16:                
                while ((UINT16)u64UserTestPattern)
                {
                    pErrorMsg = verifyMemoryBlock(currentAddress, 
                                                  (currentAddress + testBlockSize),
                                                  (UINT16)u64UserTestPattern);                 

                    if (pErrorMsg != NO_ERROR_MSG)
                    {
                        initProgress(0);
                        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
                    }
                    else
                    {
                        u64UserTestPattern <<= 1;
                    }
                }                
                break;
            case 32:
                while ((UINT32)u64UserTestPattern)
                {
                    pErrorMsg = verifyMemoryBlock(currentAddress, 
                                                  (currentAddress + testBlockSize),
                                                  (UINT32)u64UserTestPattern);                 

                    if (pErrorMsg != NO_ERROR_MSG)
                    {
                        initProgress(0);
                        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
                    }
                    else
                    {
                        u64UserTestPattern <<= 1;
                    }
                }               
                break;
            case 64:
                while (u64UserTestPattern)
                {
                    pErrorMsg = verifyMemoryBlock(currentAddress, 
                                                  (currentAddress + testBlockSize),
                                                  u64UserTestPattern);                 

                    if (pErrorMsg != NO_ERROR_MSG)
                    {
                        initProgress(0);
                        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
                    }
                    else
                    {
                        u64UserTestPattern <<= 1;
                    }                   
                }

                break;
        }        

        currentAddress += testBlockSize;
        verifiedMemorySize += testBlockSize;

        updateProgress(testBlockSize);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::dataLineTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                      PHYSICAL_ADDRESS_T in_endAddress,
///                                      UINT64             in_u64TestSize,
///                                      INT32              in_s32Controller,
///                                      INT32              in_s32Channel,
///                                      INT32              in_s32Rank,
///                                      INT32              in_s32DataLineAmount)
/// @brief      Verify memory data lines
/// @param      in_startAddress is a physical address for start address
/// @param      in_u64TestSize is an UINT64 value for test size
/// @param      in_s32Controller is an INT32 value for target controller
/// @param      in_s32Channel is an INT32 value for target channel
/// @param      in_s32Rank is an INT32 value for target rank
/// @param      in_s32DataLineAmount is 32bits value for the amount of data lines (ex: 8/16/32/64)
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::dataLineTest(PHYSICAL_ADDRESS_T in_startAddress,
                         UINT64             in_u64TestSize,
                         INT32              in_s32Controller,
                         INT32              in_s32Channel,
                         INT32              in_s32Rank,
                         INT32              in_s32DataLineAmount)
{
    ErrorMsg            *pErrorMsg = NO_ERROR_MSG;
    PHYSICAL_ADDRESS_T  rankOffset = 0, lastRankOffset = 0, currentAddress = 0;
    UINT64              u64DataByteWidth;
    UINT64              u64TestStep = 0;
    UINT64              u64UserTestPattern = 0;
    BOOL                bCompareFail = FALSE;
    string              strErrorMsg;

    switch (in_s32DataLineAmount)
    {
        case 8:
        case 16:
        case 32:
        case 64:
            // Round up memory address for alignment
            in_startAddress = alignMemoryAddress(in_startAddress, in_s32DataLineAmount);
            u64DataByteWidth = in_s32DataLineAmount >> 3;
            u64TestStep = in_u64TestSize / u64DataByteWidth;
            // Round up test size for alignment
            in_u64TestSize = u64TestStep * u64DataByteWidth;
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_DATA_LINE_TEST | ERROR_BAD_PARAMETER_THREE),
                                  "invalid data line amount",
                                  NULL);
    }

    // Get the rank offset based on the start address
    if ((pErrorMsg = m_pMemoryDDAPI->getRankOffset(in_startAddress, &rankOffset)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
    }

    lastRankOffset = rankOffset + (PHYSICAL_ADDRESS_T)in_u64TestSize;

    if (u64TestStep == 0)
    {
        return NO_ERROR_MSG;
    }

    initProgress(u64TestStep);

    while (rankOffset < lastRankOffset)
    {
        bCompareFail = FALSE;

        // Get physical address for target rank with rank offset
        pErrorMsg = m_pMemoryDDAPI->getPhysicalAddress(rankOffset, in_s32Controller, in_s32Channel, in_s32Rank, &currentAddress);
        if (pErrorMsg != NO_ERROR_MSG)
        {
            printf("0x%llx %d %d %d\n", rankOffset, in_s32Controller, in_s32Channel, in_s32Rank);
            return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
        }

        // Write, read and verify test pattern for test block size
        switch (in_s32DataLineAmount)
        {
            case 8:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)u64UserTestPattern);
                break;
            case 16:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)u64UserTestPattern);
                break;
            case 32:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)u64UserTestPattern);
                break;
            case 64:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, u64UserTestPattern);
                break;
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
        }
        else
        {
            switch (in_s32DataLineAmount)
            {
                case 8:
                    if ((UINT8)u64UserTestPattern != m_pMemoryDDAPI->read8(currentAddress))
                        bCompareFail = TRUE;
                    break;
                case 16:
                    if ((UINT16)u64UserTestPattern != m_pMemoryDDAPI->read16(currentAddress))
                        bCompareFail = TRUE;
                    break;
                case 32:
                    if ((UINT32)u64UserTestPattern != m_pMemoryDDAPI->read32(currentAddress))
                        bCompareFail = TRUE;
                    break;
                case 64:
                    if (u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress))
                        bCompareFail = TRUE;
                    break;
            }

            if (bCompareFail)
            {
                initProgress(0);
                strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, strErrorMsg.c_str());
            }
        }

        u64UserTestPattern = 1;

        switch (in_s32DataLineAmount)
        {
            case 8:
                while ((UINT8)u64UserTestPattern)
                {
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)u64UserTestPattern);

                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        if ((UINT8)u64UserTestPattern != m_pMemoryDDAPI->read8(currentAddress))
                        {
                            initProgress(0);
                            strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                            return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, strErrorMsg.c_str());
                        }
                    }
                    else
                    {
                        initProgress(0);
                        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
                    }

                    u64UserTestPattern <<= 1;
                }
                break;
            case 16:
                while ((UINT16)u64UserTestPattern)
                {
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)u64UserTestPattern);

                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        if ((UINT16)u64UserTestPattern != m_pMemoryDDAPI->read16(currentAddress))
                        {
                            initProgress(0);
                            strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                            return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, strErrorMsg.c_str());
                        }
                    }
                    else
                    {
                        initProgress(0);
                        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
                    }

                    u64UserTestPattern <<= 1;
                }
                break;
            case 32:
                while ((UINT32)u64UserTestPattern)
                {
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)u64UserTestPattern);

                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        if ((UINT32)u64UserTestPattern != m_pMemoryDDAPI->read32(currentAddress))
                        {
                            initProgress(0);
                            strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                            return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, strErrorMsg.c_str());
                        }
                    }
                    else
                    {
                        initProgress(0);
                        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
                    }

                    u64UserTestPattern <<= 1;
                }
                break;
            case 64:
                while (u64UserTestPattern)
                {
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, u64UserTestPattern);

                    if (pErrorMsg == NO_ERROR_MSG)
                    {
                        if (u64UserTestPattern != m_pMemoryDDAPI->read64(currentAddress))
                        {
                            initProgress(0);
                            strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress);
                            return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, strErrorMsg.c_str());
                        }
                    }
                    else
                    {
                        initProgress(0);
                        return returnErrorMsg(ERROR_MEMORY_DAPI_DATA_LINE_TEST, NULL, pErrorMsg);
                    }

                    u64UserTestPattern <<= 1;
                }
                break;
        }

        rankOffset += in_s32DataLineAmount >> 3;

        updateProgress(1);
    }

    return NO_ERROR_MSG;
}


