/// @file       MemoryDAPI_RandomAddressTest.cpp
/// @brief      Provide random address line test for MemoryDAPI
/// @details    Verify memory by accessing random addresses for a specified test time
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include <time.h>

#include "MemoryDAPI.hpp"
#include "CommonFunctions.hpp"

/// @fn         MemoryDAPI::randomAddressTest(PHYSICAL_ADDRESS_T    in_startAddress,
///                                           PHYSICAL_ADDRESS_T    in_endAddress,
///                                           UINT32                in_u32TestTime)
/// @brief      Verify memory address lines
/// @param      in_startAddress is a physcial address for start address
/// @param      in_endAddress is a physcial address for end address
/// @param      in_u32TestTime is 32 bits value for testing time
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::randomAddressTest(PHYSICAL_ADDRESS_T    in_startAddress,
                              PHYSICAL_ADDRESS_T    in_endAddress,
                              UINT32                in_u32TestTime)
{
    ErrorMsg*           pErrorMsg = NULL;
    string              strErrorMsg;
    PHYSICAL_ADDRESS_T  testBlockSize = 0;
    PHYSICAL_ADDRESS_T  tempOffset = 0;
    PHYSICAL_ADDRESS_T  tempAddress = 0;
    PHYSICAL_ADDRESS_T  testBlock = 0;
    UINT8               u8TestPattern = 0;
    BOOL                bInverseAddressMode = FALSE;
    double              dTimeInterval = (double)in_u32TestTime;
    time_t              startTime;
    time_t              currentTime;
    time_t              previousTime;

    testBlockSize = in_endAddress - in_startAddress;

    if ((testBlockSize == 0) || (in_u32TestTime == 0))
    {
        return NO_ERROR_MSG;
    }

    srand(time(NULL));

    initProgress(in_u32TestTime);
    
    startTime = time(NULL);
    previousTime = startTime;

    do
    {
        testBlock = (PHYSICAL_ADDRESS_T)m_pMemoryDDAPI->getMemoryMapSize();        
                
        // Generate random test pattern and random address
        u8TestPattern = rand() % 0x100;
        
        tempOffset = (((UINT64)rand()) << 32) | (UINT64)rand();

        tempAddress = in_startAddress + (tempOffset % testBlockSize);

        bInverseAddressMode = (BOOL)(rand() % 2);
        
        if ((tempAddress + testBlock) > in_endAddress)
        {
            testBlock = in_endAddress - tempAddress;
        }        

        if (bInverseAddressMode)
        {
            pErrorMsg = verifyMemoryBlock(tempAddress + testBlock, tempAddress, u8TestPattern);
        }
        else
        {
            pErrorMsg = verifyMemoryBlock(tempAddress, tempAddress + testBlock, u8TestPattern);
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_RANDOM_ADDRESS_TEST, NULL, pErrorMsg);
        }
               
        currentTime = time(NULL);

        updateProgress((UINT64)difftime(currentTime, previousTime));

        previousTime = currentTime;
    } while (difftime(currentTime, startTime) < dTimeInterval);
      
    return NO_ERROR_MSG;
}


