/// @file       NVSRAMDAPI.cpp
/// @brief      NVSRAMDAPI class
/// @details    Provide diagnostic and utility APIs for NVSRAM module
/// @version    1.0.0
/// @date       2015.11.30
/// @author     JerryLee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include <string>
#include <fstream>
#include "CommonFunctions.hpp"
#include "NVSRAMDAPI.hpp"

/// @fn         NVSRAMDAPI::NVSRAMDAPI(STRING className) : BaseClassDAPI(className)
/// @brief      NVSRAMDAPI constructor
/// @param      className is a char string for class name
/// @exception  Nothing
/// @return     Nothing
NVSRAMDAPI::NVSRAMDAPI(STRING className) : BaseClassDAPI(className)
{
    if((memDAPI = new MemoryDAPI("MEMORY_DAPI")) == NULL)
    {
        addErrorMsg( (ERRORCODE)(ERROR_NVSRAM_DAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_API_CONSTRUCTOR1_ERROR), "Create MEMORY_DAPI fail");
    }
    if((nvsramDDAPI = new NVSRAMDDAPI("NVSRAM_DRVR")) == NULL)
    {
        addErrorMsg( (ERRORCODE)(ERROR_NVSRAM_DAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_API_CONSTRUCTOR2_ERROR), "Create NVSRAM_DDAPI fail");
    }

    bufferBackup = (SYSTEM_ADDRESS_T)NULL;
    nvsramBackup = (SYSTEM_ADDRESS_T)NULL;
    testStartAddress = (PHYSICAL_ADDRESS_T)NULL;
    testEndAddress = (PHYSICAL_ADDRESS_T)NULL;
}

/// @fn         NVSRAMDAPI::~NVSRAMDAPI(void)
/// @brief      NVSRAMDAPI destructor
/// @exception  Nothing
/// @return     Nothing
NVSRAMDAPI::~NVSRAMDAPI(void)
{
    // delete what we create
    if (memDAPI != NULL)
        delete memDAPI;

    if (nvsramDDAPI != NULL)
        delete nvsramDDAPI;
}

/// @fn         NVSRAMDAPI::getInformationList(vector<NVSRAM_RESOURCE_INFO>* returnedList)
/// @brief      Get NVSRAM information list
/// @param      pointer to NVSRAM information list on the host
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVSRAMDAPI::getInformationList(vector<NVSRAM_RESOURCE_INFO>* returnedList)
{
    ErrorMsg* errorMsg = NULL;
    errorMsg = nvsramDDAPI->getDeviceList(returnedList);

    if (returnedList->empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVSRAM_DAPI_GET_INFORMATION | ERROR_NVSRAM_DAPI_NO_NVSRAM_DEVICE),
                              NULL, errorMsg);
    }
    return NO_ERROR_MSG;
}

/// @fn         NVSRAMDAPI::getInformationByID(UINT8 logicalID, NvsramInfo* returnInfo)
/// @brief      Get NVSRAM information by logicalID
/// @param      pointer to specific NVSRAM information
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
NVSRAMDAPI::getInformationByID(UINT8 logicalID, NvsramInfo* returnInfo)
{
    ErrorMsg* errorMsg = NULL;
    vector<NvsramInfo> nvsramList;

    errorMsg = nvsramDDAPI->getDeviceList(&nvsramList);

    if (nvsramList.empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVSRAM_DAPI_GET_INFORMATION | ERROR_NVSRAM_DAPI_NO_NVSRAM_DEVICE),NULL, errorMsg);
    }
    if(logicalID > nvsramList.size())
    {
        return returnErrorMsg(ERRORCODE(ERROR_NVSRAM_DAPI_GET_INFORMATION | ERROR_BAD_PARAMETER),
                              "Bad NVSRAM logical ID");
    }
    for(vector<NvsramInfo>::iterator iter = nvsramList.begin(); iter != nvsramList.end(); iter++)
    {
        if(logicalID == (iter - nvsramList.begin()))
        {
            *returnInfo = *iter;
        }
    }
    return returnErrorMsg((ERRORCODE) (ERROR_NVSRAM_DAPI_GET_INFORMATION), NULL, errorMsg);
}

