/// @file       PSUDAPI.hpp
/// @brief      PSUDAPI class
/// @details    Provide diagnostic and utility APIs for PSU module
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#include <unistd.h>
#include <stdlib.h>

#include "PSUDAPI.hpp"
#include "MemoryDDAPI.hpp"
#include "CommonFunctions.hpp"

/// @fn         PSUDAPI::PSUDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
/// @brief      PSUDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
PSUDAPI::PSUDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
{
    ErrorMsg *pErrorMsg = NO_ERROR_MSG;

    if ((m_pPSUDDAPI = new PSUDDAPI()) == NULL)
    {
        addErrorMsg(ERROR_PSU_DAPI_CONSTRUCTOR, "Fail to create PSUDDAPI");
        return;
    }

    if ((pErrorMsg = m_pPSUDDAPI->getErrorMsg()) != NO_ERROR_MSG)
    {
        addErrorMsg(ERROR_PSU_DAPI_CONSTRUCTOR, NULL, pErrorMsg);
    }
}

/// @fn         PSUDAPI::~PSUDAPI(void)
/// @brief      PSUDAPI destructor
/// @exception  Nothing
/// @return     Nothing
PSUDAPI::~PSUDAPI(void)
{
    if (m_pPSUDDAPI)
    {
        delete m_pPSUDDAPI;
        m_pPSUDDAPI = NULL;
    }
}

/// @fn         PSUDAPI::getDeviceList(string        in_strDeviceType,
///                                    list<string>  *out_pList)
/// @brief      Get the list of PSU devices
/// @param      out_pPSUList is a pointer to a list for PSU devices
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PSUDAPI::getDeviceList(string        in_strDeviceType,
                       list<string>  *out_pList)
{
    ErrorMsg*  pErrorMsg = NO_ERROR_MSG;

    if (in_strDeviceType == "psu")
    {
        if ((pErrorMsg = m_pPSUDDAPI->getDeviceList(out_pList)) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_PSU_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
        }
    }
    else if (in_strDeviceType == "fan")
    {
        if ((pErrorMsg = m_pPSUDDAPI->getFanList(out_pList)) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_PSU_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
        }
    }
    else if (in_strDeviceType == "led")
    {
        if ((pErrorMsg = m_pPSUDDAPI->getLEDList(out_pList)) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_PSU_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
        }
    }
    else if (in_strDeviceType == "temp")
    {
        if ((pErrorMsg = m_pPSUDDAPI->getTemperatureList(out_pList)) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_PSU_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
        }
    }
    else if (in_strDeviceType == "threshold")
    {
        if ((pErrorMsg = m_pPSUDDAPI->getTemperatureThresholdList(out_pList)) != NO_ERROR_MSG)
        {
            return returnErrorMsg(ERROR_PSU_DAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
        }
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DAPI_GET_DEVICE_LIST | ERROR_ILLEGAL_REQUEST));
    }
    return NO_ERROR_MSG;
}

/// @fn         PSUDAPI::getDataByElement(PSU_ELEMENT_TYPE  in_eElementType,
///                                       string            in_strDevice,
///                                       string            *out_strpValue)
/// @brief      Get data from the given device
/// @param      in_eElementType is one of PSU_ELEMENT_TYPE
/// @param      in_strDevice is a string for the PSU device
/// @param      out_strpValue is a pointer to a string for data value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PSUDAPI::getDataByElement(PSU_ELEMENT_TYPE  in_eElementType,
                          string            in_strDevice,
                          string            *out_strpValue)
{
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    INT32       s32Value;

    switch (in_eElementType)
    {
        case PSU_FW:
            pErrorMsg = m_pPSUDDAPI->getFWVersion(in_strDevice, out_strpValue);
            break;
        case PSU_MAX_POWER:
            pErrorMsg = m_pPSUDDAPI->getMaxPower(in_strDevice, out_strpValue);
            break;
        case PSU_PART_NUMBER:
            pErrorMsg = m_pPSUDDAPI->getPartNumber(in_strDevice, out_strpValue);
            break;
        case PSU_SERIAL_NUMBER:
            pErrorMsg = m_pPSUDDAPI->getSerialNumber(in_strDevice, out_strpValue);
            break;
        case PSU_VENDOR:
            pErrorMsg = m_pPSUDDAPI->getVendor(in_strDevice, out_strpValue);
            break;
        case PSU_FAN:
            pErrorMsg = m_pPSUDDAPI->getFanSpeed(in_strDevice, &s32Value);
            *out_strpValue = std::to_string(s32Value);
            break;
        case PSU_LED:
            pErrorMsg = m_pPSUDDAPI->getLEDStatus(in_strDevice, out_strpValue);
            break;
        case PSU_TEMPERATURE:
            pErrorMsg = m_pPSUDDAPI->getTemperature(in_strDevice, out_strpValue);
            break;
        case PSU_TEMPERATURE_THRESHOLD:
            pErrorMsg = m_pPSUDDAPI->getTemperatureThreshold(in_strDevice, out_strpValue);
            break;
        case PSU_TYPE:
            pErrorMsg = m_pPSUDDAPI->getType(in_strDevice, out_strpValue);
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_PSU_DAPI_GET_DATA_BY_ELEMENT | ERROR_FUNCTION_NOT_SUPPORTED));
    }

    if (pErrorMsg != NO_ERROR_MSG)
    {
        out_strpValue->clear();
        return returnErrorMsg(ERROR_PSU_DAPI_GET_DATA_BY_ELEMENT, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         PSUDAPI::getSupportValueByElement(PSU_ELEMENT_TYPE  in_eElementType,
///                                               string            in_strDevice,
///                                               list<string>      *out_pList)
/// @brief      Get support values for the given data element of the given PSU device
/// @param      in_eElementType is one of PSU_ELEMENT_TYPE
/// @param      in_strDevice is a string for the PSU device
/// @param      out_pList is a pointer to a list of strings for supported data value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PSUDAPI::getSupportValueByElement(PSU_ELEMENT_TYPE  in_eElementType,
                                  string            in_strDevice,
                                  list<string>      *out_pList)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;

    switch (in_eElementType)
    {
        case PSU_FAN:
            pErrorMsg = m_pPSUDDAPI->getSupportFanSpeed(in_strDevice, out_pList);
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_PSU_DAPI_GET_SUPPORT_VALUE_BY_ELEMENT | ERROR_FUNCTION_NOT_SUPPORTED));
    }

    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DAPI_GET_SUPPORT_VALUE_BY_ELEMENT, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         PSUDAPI::setDataByElement(PSU_ELEMENT_TYPE  in_eElementType,
///                                       string            in_strDevice,
///                                       string            in_strValue)
/// @brief      Set data to the given device
/// @param      in_eElementType is one of PSU_ELEMENT_TYPE
/// @param      in_strDevice is a string for the PSU device
/// @param      in_strValue is a string for data value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PSUDAPI::setDataByElement(PSU_ELEMENT_TYPE  in_eElementType,
                          string            in_strDevice,
                          string            in_strValue)
{
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    INT32       s32Value;

    switch (in_eElementType)
    {
        case PSU_LED:
            pErrorMsg = m_pPSUDDAPI->setLEDStatus(in_strDevice, in_strValue);
            break;
        case PSU_TEMPERATURE_THRESHOLD:
            pErrorMsg = m_pPSUDDAPI->setTemperatureThreshold(in_strDevice, in_strValue);
            break;
        case PSU_FAN:
            s32Value = std::stoi(in_strValue, NULL, 0);
            pErrorMsg = m_pPSUDDAPI->setFanSpeed(in_strDevice, s32Value);
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_PSU_DAPI_SET_DATA_BY_ELEMENT | ERROR_FUNCTION_NOT_SUPPORTED));
    }

    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DAPI_SET_DATA_BY_ELEMENT, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         PSUDAPI::updateFWByFile(string in_strDevice,
///                                     UINT8  in_u8Region,
///                                     string in_strFile)
/// @brief      Update FW to PSU
/// @param      in_strDevice is a string for the PSU device
/// @param      in_u8Region is a UINT8 for the FW region which will be updated
/// @param      in_strFile is a string for FW's file path
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PSUDAPI::updateFWByFile(string in_strDevice,
                        UINT8  in_u8Region,
                        string in_strFile)
{
    FILE        *pFile = NULL;
    INT32       s32FileSize = 0;
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    MemoryDDAPI memoryDDAPI("MemoryDDAPI");
    SYSTEM_ADDRESS_T fwBuffer;

    if ((pErrorMsg = memoryDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DAPI_UPDATE_FW_BY_FILE, NULL, pErrorMsg);
    }

    if ((pFile = fopen(in_strFile.c_str(), "r")) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DAPI_UPDATE_FW_BY_FILE | ERROR_OPEN_FILE),
                              in_strFile.c_str());
    }

    // Allocate a memory buffer for FW
    fseek(pFile, 0, SEEK_END);
    s32FileSize = ftell(pFile);
    fseek(pFile, 0, SEEK_SET);

    if ((pErrorMsg = memoryDDAPI.getMemoryBuffer(s32FileSize, &fwBuffer)) != NO_ERROR_MSG)
    {
        fclose(pFile);
        return returnErrorMsg(ERROR_PSU_DDAPI_PROGRAM, NULL, pErrorMsg);
    }

    if (s32FileSize != (INT32)fread((CHAR *)fwBuffer, 1, s32FileSize, pFile))
    {
        fclose(pFile);
        memoryDDAPI.releaseMemoryBuffer(fwBuffer);
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DAPI_UPDATE_FW_BY_FILE | ERROR_READ_FILE_FAIL),
                              in_strFile.c_str());
    }

    fclose(pFile);

    pErrorMsg = m_pPSUDDAPI->program(in_strDevice, in_u8Region, (CHAR *)fwBuffer, s32FileSize);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        memoryDDAPI.releaseMemoryBuffer(fwBuffer);
        return returnErrorMsg(ERROR_PSU_DAPI_UPDATE_FW_BY_FILE, NULL, pErrorMsg);
    }

    memoryDDAPI.releaseMemoryBuffer(fwBuffer);

    return NO_ERROR_MSG;
}
