/// @file       SASDAPI.cpp
/// @brief      SASDAPI class
/// @details    Provide diagnostic and utility APIs for SAS module
/// @version    1.0.0
/// @date       2014.01.06
/// @author     Jerry Lee
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "SASDAPI.hpp"
#include "CommonFunctions.hpp"
#include <string.h>
#include <unistd.h>
#include <iomanip>
#include <thread>
/// @fn         SASDAPI::SASDAPI(STRING in_szClassName) : BaseClassDAPI(in_szClassName)
/// @brief      SASDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
SASDAPI::SASDAPI(STRING className) : BaseClassDAPI(className)
{
    // Initializations.
    if((sasDDAPI = new SASDDAPI("SAS_DRVR")) == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_SAS_DAPI_ERRORCODE | ERROR_DIAGNOSTIC_API_CONSTRUCTOR2_ERROR), "Create SAS DDAPI fail");
    }

    memDDAPI = NULL;
    scsiFactory = NULL;
    dataBuffer = (UINT32*)NULL;
    logicalSectorToTest = 0;
    rwBlocks = 0;
    testSize = 0;
    transferSize = 0;
    throttle = 0;
    testPattern = 0;
    testSpeed = 0;
};

/// @fn         SASDAPI::~SASDAPI(void)
/// @brief      SASDAPI destructor
/// @exception  Nothing
/// @return     Nothing
SASDAPI::~SASDAPI(void)
{
    // delete SAS Driver API
    if(sasDDAPI != NULL)
        delete sasDDAPI;
};

/// @fn         SASDAPI::clearPHYErrorCountByController(UINT8 controllerID)
/// @brief      call driver ioctl to clean all PHY's error count of IOC
/// @param      IOC ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::clearPHYErrorCountByController(UINT8 controllerID)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::CLEAR_ERR_LOG, NULL);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_CLEAR_PHY_ERROR_LOG), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::clearPHYErrorCountByPHY(UINT8 controllerID, UINT8 phyNumber)
/// @brief      call driver ioctl to clean specific PHY's error count
/// @param      IOC ID
/// @param      PHY number
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::clearPHYErrorCountByPHY(UINT8 controllerID, UINT8 phyNumber)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);
    sasDDAPI->setPHYNumber(phyNumber);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::CLEAR_ERR_LOG_BY_PHY, NULL);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_CLEAR_PHY_ERROR_LOG), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getAddressByController(UINT8 controllerID, UINT64* returnAddress)
/// @brief      call driver ioctl to get specific controller SAS address
/// @param      IOC ID
/// @param      A pointer to the return SAS address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getAddressByController(UINT8 controllerID, UINT64* returnAddress)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;
    //vector<SAS_CONTROLLER_INFO> sasList;
    //string pciBDF;
    //CHAR* scsiDirToRead = (CHAR *)"/sys/class/scsi_host/";
    //DIR* dir = opendir(scsiDirToRead);
    //struct dirent *dirent;
    //struct stat fileAttribute;
    //CHAR tempBuffer[256];
    //CHAR fileName[256] = "";
    UINT32 tempAddr[2];

    sasDDAPI->setControllerID(controllerID);

    errorMsg = sasDDAPI->doIOCTLCommand( SASIOCTL::IOC_SAS_ADDRESS, &tempAddr);
    *returnAddress = tempAddr[0];
    *returnAddress = (*returnAddress << 32) + tempAddr[1];

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_TEMPERATURE), NULL, errorMsg);
    }
/*
    errorMsg = sasDDAPI->getControllerList(&sasList);

    if(sasList.empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_HOST_SAS_ADDRESS), NULL, errorMsg);
    }

    for(vector<SAS_CONTROLLER_INFO>::iterator iter = sasList.begin(); iter != sasList.end(); iter++)
    {
        if (controllerID == (iter - sasList.begin()))
        {
            pciBDF = iter->domindBusDeviceFunction;

            while((dirent = readdir(dir)) != NULL)
            {
                snprintf(fileName, sizeof(fileName), "%s%s", scsiDirToRead, dirent->d_name);
                lstat(fileName, &fileAttribute);

                if(S_ISDIR(fileAttribute.st_mode))
                {
                    // This directory dirent is another directory
                }
                else if(S_ISLNK(fileAttribute.st_mode))
                {
                    INT32 linkLength = readlink(fileName, tempBuffer, sizeof(tempBuffer) - 1);
                    if(linkLength != -1)
                    {
                        tempBuffer[linkLength] = '\0';
                    }
                    else
                    {
                        closedir(dir);
                        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_HOST_SAS_ADDRESS | ERROR_SAS_DDAPI_READ_LINK_ERROR), NULL, NULL);;
                    }

                    CHAR tempString[256] = "";
                    snprintf(tempString, sizeof(tempString), "%s", pciBDF.c_str());
                    CHAR *chpAddress = strstr(tempBuffer, tempString);

                    if(chpAddress == NULL)
                    {
                        continue;
                    }

                    // start to get SAS address
                    CHAR devicePath[128] = "", tempAddress[20];
                    FILE *fp = NULL;
                    snprintf(devicePath, sizeof(devicePath), "cat /sys/class/scsi_host/%s/host_sas_address", dirent->d_name);
                    fp = popen(devicePath, "r");
                    if(fgets(tempAddress, sizeof(tempAddress), fp) != NULL)
                    {
                        if (tempAddress[strlen(tempAddress)-1] == '\n')
                            tempAddress[strlen(tempAddress)-1] = '\0';
                        *returnAddress = std::strtoull(tempAddress, NULL, 16);
                    }
                    if(fp)
                        pclose(fp);
                }
                else if(S_ISREG(fileAttribute.st_mode))
                {
                  // This dirent is a regular filename
                }
            }
            closedir(dir);
            break;
        }
    }
*/
    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getAttachDeviceByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnType)
/// @brief      call driver ioctl to display attach device
/// @param      IOC ID
/// @param      PHY number
/// @param      return attached type
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getAttachDeviceByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnType)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);
    sasDDAPI->setPHYNumber(phyNumber);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::SHOW_ATTACHED_DEVICE, returnType);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_ATTACHED_DEVICE), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getDriveDeviceNodeListByController(UINT8 controllerID, vector<string>* nodeList)
/// @brief      call driver ioctl to clean specific PHY's error count
/// @param      IOC ID
/// @param      A pointer to the return device node list
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getDriveDeviceNodeListByController(UINT8 controllerID, vector<SAS_DRIVE_INFO>* nodeList)
{
    ErrorMsg *errorMsg = NO_ERROR_MSG;
    vector<SAS_CONTROLLER_INFO> sasList;

    errorMsg = sasDDAPI->getControllerList(&sasList);

    if(sasList.empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_INFORMATION), NULL, errorMsg);
    }
    if(controllerID > sasList.size())
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_INFORMATION | ERROR_BAD_PARAMETER),
                              "Bad IOC logical ID");
    }
    if((errorMsg = sasDDAPI->getDriveListByController(controllerID, nodeList)) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_SAS_DAPI_GET_DEVICE_NODE_LIST, NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getErrorCountByPHY(UINT8 controllerID, UINT8 phyNumber, ErrorCountInfo* returnInfo)
/// @brief      call driver ioctl to display attach device
/// @param      IOC ID
/// @param      PHY number
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getErrorCountByPHY(UINT8 controllerID, UINT8 phyNumber, ErrorCountInfo* returnInfo)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);
    sasDDAPI->setPHYNumber(phyNumber);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::PHY_ERROR_COUNT, returnInfo);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_ERROR_COUNT), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         getErrorCountByController(UINT8 controllerID, ErrorCountInfo* returnInfo)
/// @brief      call driver ioctl to display attach device
/// @param      IOC ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getErrorCountByController(UINT8 controllerID, ErrorCountInfo* returnInfo)
{
    ErrorMsg* errorMsg = NULL;

    sasDDAPI->setControllerID(controllerID);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::PHY_ERROR_COUNT, returnInfo);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_ERROR_COUNT), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getInformationByController(UINT8 controllerID, SasInfo* returnInfo)
/// @brief      call driver ioctl to get SAS controller basic informations
/// @param      IOC ID
/// @param      pointer to return SAS Info structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getInformationByController(UINT8 controllerID, SasInfo* returnInfo)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;
    vector<SAS_CONTROLLER_INFO> sasList;

    errorMsg = sasDDAPI->getControllerList(&sasList);

    if(sasList.empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_INFORMATION), NULL, errorMsg);
    }
    if(controllerID > sasList.size())
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_INFORMATION | ERROR_BAD_PARAMETER),
                              "Bad IOC logical ID");
    }

    errorMsg = sasDDAPI->getInformationByController(controllerID, returnInfo);

    if(errorMsg)
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_INFORMATION), NULL, errorMsg);

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getInformationByDriveName(string wellKnownName, SasDriveInfo* returnInfo)
/// @brief      call driver ioctl to get SAS drive informations
/// @param      SAS drive well known name
/// @param      pointer to return SAS drive info structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getInformationByDriveName(string wellKnownName, SasDriveInfo* returnInfo)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;
    vector<SAS_CONTROLLER_INFO> sasList;
    vector<SAS_DRIVE_INFO> driveList;

    getInformationList(&sasList);

    for(vector<SAS_CONTROLLER_INFO>::iterator sasIter = sasList.begin(); sasIter != sasList.end(); ++sasIter)
    {
        errorMsg = getDriveDeviceNodeListByController((sasIter - sasList.begin()), &driveList);
        if (driveList.empty())
            continue;
        else
        {
            for(vector<SAS_DRIVE_INFO>::iterator driveIter = driveList.begin(); driveIter != driveList.end(); ++driveIter)
            {
                if(wellKnownName == driveIter->wellKnownName)
                {
                    returnInfo->wellKnownName = driveIter->wellKnownName;
                    returnInfo->controllerName = driveIter->controllerName;
                    returnInfo->deviceNodeName = driveIter->deviceNodeName;
                    returnInfo->sgDeviceNodeName = driveIter->sgDeviceNodeName;
                    returnInfo->sasAddress = driveIter->sasAddress;
                    errorMsg = sasDDAPI->getDriveInfoByName(driveIter->wellKnownName, returnInfo);

                    if(errorMsg != NO_ERROR_MSG)
                        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_DRIVE_INFORMATION), NULL, errorMsg);
                    else
                        return NO_ERROR_MSG;
                }
            }
        }
    }

    return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_DRIVE_INFORMATION), "NO Attached SAS Drive", NULL);
}

/// @fn         SASDAPI::getInformationList(vector<SAS_CONTROLLER_INFO>* returnedList)
/// @brief      Get SAS information list
/// @param      pointer to SAS information list on the host
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getInformationList(vector<SAS_CONTROLLER_INFO>* returnList)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    errorMsg = sasDDAPI->getControllerList(returnList);

    if(returnList->empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_INFORMATION), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getLinkSpeedByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnSpeed)
/// @brief      call driver ioctl to display attach device
/// @param      IOC ID
/// @param      PHY number
/// @param      return attached type
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getLinkSpeedByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnSpeed)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);
    sasDDAPI->setPHYNumber(phyNumber);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::LINK_SPEED, returnSpeed);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_LINK_SPEED), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getLinkStatusByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnType)
/// @brief      call driver ioctl to display attach device
/// @param      IOC ID
/// @param      PHY number
/// @param      return attached type
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getLinkStatusByPHY(UINT8 controllerID, UINT8 phyNumber, string* returnStatus)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);
    sasDDAPI->setPHYNumber(phyNumber);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::LINK_STATUS, returnStatus);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_LINK_STATUS), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::getLogicalIDByName(string wellKnowName, UINT8* returnID)
/// @brief      Get SAS controller logical ID
/// @param      Controller well known name
/// @param      pointer to SAS controller logical ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getLogicalIDByName(string wellKnowName, UINT8* returnID)
{
    vector<SAS_CONTROLLER_INFO> sasList;
    ErrorMsg* errorMsg = NO_ERROR_MSG;
    errorMsg = sasDDAPI->getControllerList(&sasList);

    if(sasList.empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_LOGICALID_BY_NAME), NULL, errorMsg);
    }
    for(vector<SAS_CONTROLLER_INFO>::iterator iter = sasList.begin(); iter != sasList.end(); iter++)
    {
        if (wellKnowName == iter->wellKnowName)
        {
            *returnID = (iter - sasList.begin());
            return NO_ERROR_MSG;
        }
    }

    return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_LOGICALID_BY_NAME | ERROR_BAD_PARAMETER_FIVE),
                                    NULL, NULL);
}

/// @fn         SASDAPI::getSmartDataByDriveName(string wellKnownName, StorageSASSMARTDataStruct* smartData)
/// @brief      call driver ioctl to get SAS drive informations
/// @param      SAS drive well known name
/// @param      pointer to return SAS drive info structure
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getSmartDataByDriveName(string wellKnownName, StorageSASSMARTDataStruct* smartData)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;
    vector<SAS_CONTROLLER_INFO> sasList;
    vector<SAS_DRIVE_INFO> driveList;

    getInformationList(&sasList);

    for(vector<SAS_CONTROLLER_INFO>::iterator sasIter = sasList.begin(); sasIter != sasList.end(); ++sasIter)
    {
        errorMsg = getDriveDeviceNodeListByController((sasIter - sasList.begin()), &driveList);
        if (driveList.empty())
            continue;
        else
        {
            for(vector<SAS_DRIVE_INFO>::iterator driveIter = driveList.begin(); driveIter != driveList.end(); ++driveIter)
            {
                if(wellKnownName == driveIter->wellKnownName)
                {
                    errorMsg = sasDDAPI->getDriveInfoByName(driveIter->wellKnownName, &(*driveIter));

                    if(errorMsg != NO_ERROR_MSG)
                        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_DRIVE_INFORMATION), NULL, errorMsg);
                    else
                    {
                        *smartData = driveIter->smartData;
                        return NO_ERROR_MSG;
                    }
                }
            }
        }
    }

    return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_DRIVE_INFORMATION), "NO Attached SAS Drive", NULL);
}

/// @fn         SASDAPI::getTemperatureByController(UINT8 controllerID, UINT32* out_Temperature)
/// @brief      call driver ioctl to get current temperature
/// @param      IOC ID
/// @param      pointer to return Temperature
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::getTemperatureByController(UINT8 controllerID, UINT32* returnTemperature)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::CURRENT_TEMP, returnTemperature);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_GET_TEMPERATURE), NULL, errorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::doIOCTL
/// @brief      Execute specific IOCLT command
/// @param      Specific IOC controller
/// @param      SAS ioctl operation code
/// @param      The returned value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::doIOCTL(UINT8 controllerID, UINT32 ioctlOperation, UINT32* returnedValue)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);

    // only support no argument input ioctl OP code
    errorMsg = sasDDAPI->doIOCTLCommand((SASIOCTL::IOCTL_OP_CODE)ioctlOperation, returnedValue);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_DO_IOCTL), NULL, errorMsg);
    }
    
    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::doLoopbackTest(UINT8 controllerID)
/// @brief      call driver ioctl to execute loopback test (normal disk I/O like test)
/// @param      IOC ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::doLoopbackTest(UINT8 controllerID)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::EXECUTE_LOOPBACK, NULL);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_LOOPBACK_TEST), NULL, errorMsg);
    }
    
    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::programSASAddress(UINT8 controllerID, string sasAddress)
/// @brief      Program specific IOC SAS address
/// @param      Specific IOC controller
/// @param      SAS address to be programmed
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::programSASAddress(UINT8 controllerID, string sasaddress)
{
    CHAR command[128] = "", resultString[100];
    string tempString;
    UINT8 lineNum = 0;
    FILE *fp = NULL;

    snprintf(command, sizeof(command), "sas3flash -o -c %d -sasadd %s -l /tmp/sasADDRflash.log", furyRealID[controllerID], sasaddress.c_str());
    fp = popen(command, "w");
    if(!fp)
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_PROGRAM_ADDRESS | ERROR_READ_FILE_FAIL), NULL, NULL);

    if(fp)
        pclose(fp);

    fp = fopen("/tmp/sasADDRflash.log", "r");
    if(!fp)
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_PROGRAM_ADDRESS | ERROR_READ_FILE_FAIL), NULL, NULL);

    while(fgets(resultString, sizeof(resultString), fp))
    {
        if(resultString[strlen(resultString)-1] == '\n')
            resultString[strlen(resultString)-1] = '\0';

        if(lineNum == 10)
        {
            if(strstr(resultString, "Finished Processing Commands Successfully"))
                break;
            else
            {
                if(fp)
                    fclose(fp);

                return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_PROGRAM_ADDRESS), "SAS address programming fail", NULL);
            }
        }
        lineNum++;
    }

    if(fp)
        fclose(fp);
    
    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::programSASFirmware(UINT8 controllerID, string filepath)
/// @brief      Program specific IOC SAS Firmware
/// @param      Specific IOC controller
/// @param      SAS Firmware file path
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::programSASFirmware(UINT8 controllerID, string filepath)
{
    CHAR command[128] = "", resultString[100];
    string tempString;
    UINT8 lineNum = 0;
    BOOL result = FALSE;
    FILE *fp = NULL;

    snprintf(command, sizeof(command), "sas3flash -o -c %d -f %s -l /tmp/sasFWflash.log", furyRealID[controllerID], filepath.c_str());

    fp = popen(command, "w");
    if(!fp)
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_PROGRAM_FIRMWARE | ERROR_READ_FILE_FAIL), NULL, NULL);

    if(fp)
        pclose(fp);

    fp = fopen("/tmp/sasFWflash.log", "r");
    if(!fp)
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_PROGRAM_ADDRESS | ERROR_READ_FILE_FAIL), NULL, NULL);

    while(fgets(resultString, sizeof(resultString), fp))
    {
        if(resultString[strlen(resultString)-1] == '\n')
            resultString[strlen(resultString)-1] = '\0';

        if(lineNum == 30)
        {
            if(strstr(resultString, "Finished Processing Commands Successfully"))
                break;
            else
            {
                if(fp)
                    fclose(fp);

                return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_PROGRAM_FIRMWARE), "SAS FW programming fail", NULL);
            }
        }
        lineNum++;
    }
    if(fp)
        pclose(fp);

    return NO_ERROR_MSG;
}

/// @fn         SASDAPI::resetIOC(UINT8 controllerID)
/// @brief      reset specific IOC controller
/// @param      IOC ID
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SASDAPI::resetIOC(UINT8 controllerID)
{
    ErrorMsg* errorMsg = NO_ERROR_MSG;

    sasDDAPI->setControllerID(controllerID);

    errorMsg = sasDDAPI->doIOCTLCommand(SASIOCTL::EXECUTE_IOC_HARD_RESET, NULL);

    if(errorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SAS_DAPI_RESETIOC), NULL, errorMsg);
    }
    
    return NO_ERROR_MSG;
}

ErrorMsg*
SASDAPI::readFromDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize)
{
    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

ErrorMsg*
SASDAPI::writeToDevice(string deviceNode, UINT32* buffer, UINT32 startLogical, UINT32 testSize)
{
    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

