/// @file       FPGADDAPI.hpp
/// @brief      FPGADDAPI class
/// @details    Provide low layer diagnostic and utility APIs for FPGA module
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#include <sys/io.h>
#include "FPGADDAPI.hpp"
#include "FPGARegConfig.hpp"
#include "MemoryDDAPI.hpp"


/// @fn         FPGADDAPI::FPGADDAPI() : BaseClassDDAPI("FPGADDAPI")
/// @brief      FPGADDAPI constructor
/// @exception  Nothing
/// @return     Nothing
FPGADDAPI::FPGADDAPI() : BaseClassDDAPI("FPGADDAPI")
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;

    m_u32BaseAddress = getBaseAddress();

    if (m_u32BaseAddress == 0)
    {
        addErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_CONSTRUCTOR | ERROR_FPGA_DDAPI_UNKNOWN_BASE_ADDRESS),
                    NULL);
        return;
    }

    if ((m_pMemoryDDAPI = new MemoryDDAPI("MemoryDDAPI")) == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_DRIVER_API_CONSTRUCTOR1_ERROR),
                    "fail to create MemoryDDAPI");
    }
    else if ((pErrorMsg = m_pMemoryDDAPI->getErrorMsg()) != NO_ERROR_MSG)
    {
        addErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_DRIVER_API_CONSTRUCTOR1_ERROR),
                    "fail to create MemoryDDAPI",
                    pErrorMsg);
    }
}

/// @fn         FPGADDAPI::~FPGADDAPI(void)
/// @brief      FPGADDAPI destructor
/// @exception  Nothing
/// @return     Nothing
FPGADDAPI::~FPGADDAPI(void)
{

}

/// @fn         FPGADDAPI::getDeviceList(list<string> *out_pList)
///
/// @brief      Get device name list
/// @param      out_pList is a pointer to a list of string for device name
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADDAPI::getDeviceList(list<string> *out_pList)
{
    if (out_pList == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_GET_DEVICE_LIST | ERROR_BAD_PARAMETER),
                              "NULL pointer");
    }

    out_pList->push_back(FPGA_DEVICE_NAME);

    return NO_ERROR_MSG;
}

/// @fn         FPGADDAPI::getSPADInfo(string in_strDeviceName,
///                                    UINT32 *out_u32pStartOffset,
///                                    UINT32 *out_u32pTotalSize)
/// @brief      Get information about scratch pad registers
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      out_u32pStartOffset is a pointer to a UINT32 value for register start offset
/// @param      out_u32pTotalSize is a pointer to a UINT32 value for total size of registers
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADDAPI::getSPADInfo(string in_strDeviceName,
                       UINT32 *out_u32pStartOffset,
                       UINT32 *out_u32pTotalSize)
{
    if (out_u32pStartOffset == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_GET_SPAD_INFO | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer");
    }

    if (out_u32pTotalSize == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_GET_SPAD_INFO | ERROR_BAD_PARAMETER_THREE),
                              "NULL pointer");
    }

    if (in_strDeviceName != FPGA_DEVICE_NAME)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_GET_SPAD_INFO | ERROR_NO_DEVICE_FOUND));
    }

    *out_u32pStartOffset = 0;
    *out_u32pTotalSize = FPGA_SPAD_REG_TOTAL_SIZE;

    return NO_ERROR_MSG;
}

/// @fn         FPGADDAPI::getRegSize(UINT32 in_u32Offset)
///
/// @brief      Get register size in bytes
/// @param      in_u32Offset is a UINT32 for register offset
/// @exception  Nothing
/// @return     A UINT32 value for register size in bytes
UINT32
FPGADDAPI::getRegSize(UINT32 in_u32Offset)
{
    INT32   s32RegCount = FPGA_REG_COUNT;
    INT32   i;

    for (i = 0; i < s32RegCount; i++)
    {
        if (in_u32Offset == regList[i].u32Offset)
        {
            return regList[i].u8Width;
        }
    }

    return 0;
}

/// @fn         FPGADDAPI::read(string  in_strDeviceName,
///                             UINT32  in_u32Offset
///                             UINT8   *out_u8pValue)
/// @brief      Read value from register
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      in_u32Offset is a UINT32 value for register offset
/// @param      out_u8pValue is a pointer to a UINT8 value for register's value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADDAPI::read(string  in_strDeviceName,
                UINT32  in_u32Offset,
                UINT8   *out_u8pValue)
{
    FPGARegInfo *pRegInfo = getRegInfo(in_u32Offset);
    string      strTemp;
    PHYSICAL_ADDRESS_T address;
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;

    if (out_u8pValue == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_READ | ERROR_BAD_PARAMETER_THREE),
                              "NULL pointer");
    }

    if (in_strDeviceName != FPGA_DEVICE_NAME)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_READ | ERROR_NO_DEVICE_FOUND));
    }

    if (pRegInfo == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_READ | ERROR_FPGA_DDAPI_UNKNOWN_REGISTER));
    }

    if (pRegInfo->s32Attribute == WRITE_ONLY)
    {
        strTemp = "register: 0x" + Int2HexStr(in_u32Offset);
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_READ | ERROR_REGISTER_IS_WRITE_ONLY),
                              strTemp.c_str());
    }

    address = (PHYSICAL_ADDRESS_T)(m_u32BaseAddress + pRegInfo->u32Offset);
    *out_u8pValue = m_pMemoryDDAPI->read8(address);

    if ((*out_u8pValue == 0xde) &&
        ((pErrorMsg = m_pMemoryDDAPI->getErrorMsg()) != NO_ERROR_MSG))
    {
        return returnErrorMsg(ERROR_FPGA_DDAPI_READ, NULL, pErrorMsg);
    }


    return NO_ERROR_MSG;
}

/// @fn         FPGADDAPI::readSPAD(string  in_strDeviceName,
///                                 UINT32  in_u32Offset
///                                 UINT8   *out_u8pValue)
/// @brief      Read value from scratch pad register
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      in_u32Offset is a UINT32 value for register offset
/// @param      out_u8pValue is a pointer to a UINT8 value for register's value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADDAPI::readSPAD(string  in_strDeviceName,
                    UINT32  in_u32Offset,
                    UINT8   *out_u8pValue)
{
    string      strTemp;
    PHYSICAL_ADDRESS_T address;

    if (out_u8pValue == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_READ_SPAD | ERROR_BAD_PARAMETER_THREE),
                              "NULL pointer");
    }

    if (in_strDeviceName != FPGA_DEVICE_NAME)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_READ_SPAD | ERROR_NO_DEVICE_FOUND));
    }

    address = (PHYSICAL_ADDRESS_T)(m_u32BaseAddress + FPGA_SPAD_REG_OFFSET_START + in_u32Offset);
    *out_u8pValue = m_pMemoryDDAPI->read8(address);

    return NO_ERROR_MSG;
}

/// @fn         FPGADDAPI::write(string  in_strDeviceName,
///                              UINT32  in_u32Offset
///                              UINT8   in_u9Value)
/// @brief      Write value to register
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      in_u32Offset is a UINT32 value for register offset
/// @param      in_u8Value is a UINT8 value for register's value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADDAPI::write(string in_strDeviceName,
                 UINT32 in_u32Offset,
                 UINT8  in_u8Value)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    string      strTemp;
    FPGARegInfo *pRegInfo = getRegInfo(in_u32Offset);
    PHYSICAL_ADDRESS_T address;

    if (in_strDeviceName != FPGA_DEVICE_NAME)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_WRITE | ERROR_NO_DEVICE_FOUND));
    }

    if (pRegInfo == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_WRITE | ERROR_FPGA_DDAPI_UNKNOWN_REGISTER));
    }

    if (pRegInfo->s32Attribute == READ_ONLY)
    {
        strTemp = "register: 0x" + Int2HexStr(in_u32Offset);
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_WRITE | ERROR_REGISTER_IS_READ_ONLY),
                              strTemp.c_str());
    }

    address = (PHYSICAL_ADDRESS_T)(m_u32BaseAddress + pRegInfo->u32Offset);
    pErrorMsg = m_pMemoryDDAPI->write(address, in_u8Value);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_FPGA_DDAPI_WRITE, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         FPGADDAPI::writeSPAD(string  in_strDeviceName,
///                                  UINT32  in_u32Offset
///                                  UINT8   in_u9Value)
/// @brief      Write value to scratch pad register
/// @param      in_strDeviceName is a string for FPGA device name
/// @param      in_u32Offset is a UINT32 value for register offset
/// @param      in_u8Value is a UINT8 value for register's value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
FPGADDAPI::writeSPAD(string in_strDeviceName,
                     UINT32 in_u32Offset,
                     UINT8  in_u8Value)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    PHYSICAL_ADDRESS_T address;

    if (in_strDeviceName != FPGA_DEVICE_NAME)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_FPGA_DDAPI_WRITE_SPAD | ERROR_NO_DEVICE_FOUND));
    }

    address = (PHYSICAL_ADDRESS_T)(m_u32BaseAddress + FPGA_SPAD_REG_OFFSET_START + in_u32Offset);
    pErrorMsg = m_pMemoryDDAPI->write(address, in_u8Value);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_FPGA_DDAPI_WRITE_SPAD, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}


/// @fn         FPGADDAPI::getBaseAddress()
///
/// @brief      Get memory mapped base address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
UINT32
FPGADDAPI::getBaseAddress(void)
{
    UINT8   u8Temp;
    UINT32  u32Address = 0;

    // Enable high IO port
    if (iopl(3) != 0)
        return 0;

    // Base address is defined in SPAR0-1
    u8Temp = inb(0x700);
    u32Address |= (UINT32)u8Temp << 16;

    u8Temp = inb(0x701);
    u32Address |= (UINT32)u8Temp << 24;

    // Disable high IO port
    iopl (0);

    return u32Address;
}

/// @fn         FPGADDAPI::getRegInfo(UINT32 in_u32Offset)
///
/// @brief      Get register's information
/// @param      in_u32Offset is a UINT32 value for register offset
/// @exception  Nothing
/// @return     A pointer to FPGARegInfo for register
FPGARegInfo *
FPGADDAPI::getRegInfo(UINT32 in_u32Offset)
{
    INT32   s32RegCount = FPGA_REG_COUNT;
    INT32   i;

    for (i = 0; i < s32RegCount; i++)
    {
        if (in_u32Offset == regList[i].u32Offset)
        {
            return &regList[i];
        }
    }

    return NULL;
}
