/// @file       PMBusPSU.cpp
/// @brief      PMBusPSU class
/// @details    Provide common data access APIs for PMBus
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "PMBusPSU.hpp"

/// @fn         PMBusPSU::PMBusPSU(void) : BaseClassPSUDDAPI()
///
/// @brief      PMBusPSU constructor
/// @exception  Nothing
/// @return     Nothing
PMBusPSU::PMBusPSU(void) : BaseClassPSUDDAPI()
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;

    if ((m_pI2CDAPI = new I2CDAPI("I2CDAPI")) == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_API_CONSTRUCTOR1_ERROR), 
                    "create I2CDAPI fail");
    }
    else if ((pErrorMsg = m_pI2CDAPI->getErrorMsg()) != NO_ERROR_MSG)
    {
        addErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_API_CONSTRUCTOR1_ERROR), 
                    "create I2CDAPI fail",
                    pErrorMsg);
    }
}

/// @fn         PMBusPSU::~PMBusPSU(void)
/// @brief      PMBusPSU destructor
/// @exception  Nothing
/// @return     Nothing
PMBusPSU::~PMBusPSU(void)
{
    if (m_pI2CDAPI)
    {
        delete m_pI2CDAPI;
    }
}

/// @fn         PMBusPSU::getFanList(list<string> *out_pList)
///
/// @brief      Get list of FAN names
/// @param      out_pList is a pointer to a list of strings for FAN names
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::getFanList(list<string> *out_pList)
{
    if (FAN1_NAME)
    {
        out_pList->push_back(FAN1_NAME);
    }

    if (FAN2_NAME)
    {
        out_pList->push_back(FAN2_NAME);
    }

    if (FAN3_NAME)
    {
        out_pList->push_back(FAN3_NAME);
    }
    
    if (FAN4_NAME)
    {
        out_pList->push_back(FAN4_NAME);
    }

    return NO_ERROR_MSG;
}

/// @fn         PMBusPSU::getFanSpeed(string  in_strDeviceName,
///                                   INT32   *out_s32pValue)
///
/// @brief      Get current speed for the given FAN
/// @param      in_strDeviceName is a string for FAN name
/// @param      out_s32pValue is a pointer to a INT32 for the current speed
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::getFanSpeed(string  in_strDeviceName,
                      INT32   *out_s32pValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    INT32       s32RegOffset = 0x0, s32Count = 0;
    UINT16      u16Data;

    if (FAN1_NAME && ((CHAR *)FAN1_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_READ_FAN_SPEED_1;
    }
    else if (FAN2_NAME && ((CHAR *)FAN2_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_READ_FAN_SPEED_2;
    }
    else if (FAN3_NAME && ((CHAR *)FAN3_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_READ_FAN_SPEED_3;
    }
    else if (FAN4_NAME && ((CHAR *)FAN4_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_READ_FAN_SPEED_4;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_GET_FAN_SPEED | ERROR_NO_DEVICE_FOUND),
                              in_strDeviceName.c_str());
    }

    pErrorMsg = m_pI2CDAPI->Read(getName(), s32RegOffset, s32RegOffset, &u16Data, &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_FAN_SPEED, NULL, pErrorMsg);
    }

    *out_s32pValue = (INT32)transformLinearToHuman(u16Data);

    return NO_ERROR_MSG;
}

/// @fn         PMBusPSU::getMaxPower(string *out_strpValue)
///
/// @brief      Get maximum power
/// @param      out_strpValue is a pointer to a string for the maximum power
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::getMaxPower(string *out_strpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT16      u16Data;
    INT32       s32Count = 0;

    pErrorMsg = m_pI2CDAPI->Read(getName(), PMBUS_CMD_MFR_POUT_MAX, PMBUS_CMD_MFR_POUT_MAX, &u16Data, &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_MAX_POWER, NULL, pErrorMsg);
    }

    *out_strpValue = std::to_string((INT32)transformLinearToHuman(u16Data));

    return NO_ERROR_MSG;
}

/// @fn         PMBusPSU::getTemperature(string   in_strDeviceName,
///                                      FLOAT    *out_fpValue)
///
/// @brief      Get current temperature
/// @param      in_strDeviceName is a string for temperature sensor name
/// @param      out_fpValue is a pointer to a FLOAT for the current temperature
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::getTemperature(string   in_strDeviceName,
                         FLOAT    *out_fpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    INT32       s32RegOffset = 0x0, s32Count = 0;
    UINT16      u16Data;

    if (TEMP1_NAME && ((CHAR *)TEMP1_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_READ_TEMPERATURE_1;
    }
    else if (TEMP2_NAME && ((CHAR *)TEMP2_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_READ_TEMPERATURE_2;
    }
    else if (TEMP3_NAME && ((CHAR *)TEMP3_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_READ_TEMPERATURE_3;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_GET_TEMPERATURE | ERROR_NO_DEVICE_FOUND),
                              in_strDeviceName.c_str());
    }

    pErrorMsg = m_pI2CDAPI->Read(getName(), s32RegOffset, s32RegOffset, &u16Data, &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_TEMPERATURE, NULL, pErrorMsg);
    }

    *out_fpValue = transformLinearToHuman(u16Data);

    return NO_ERROR_MSG;
}

/// @fn         PMBusPSU::getTemperatureList(list<string> *out_pList)
///
/// @brief      Get the list of temperature sensor names
/// @param      out_pList is a pointer to a list of strings for temperature sensor names
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::getTemperatureList(list<string> *out_pList)
{
    if (TEMP1_NAME)
    {
        out_pList->push_back(TEMP1_NAME);
    }

    if (TEMP2_NAME)
    {
        out_pList->push_back(TEMP2_NAME);
    }

    if (TEMP3_NAME)
    {
        out_pList->push_back(TEMP3_NAME);
    }

    return NO_ERROR_MSG;
}

/// @fn         PMBusPSU::getTemperatureThreshold(string   in_strDeviceName,
///                                               FLOAT    *out_fpValue)
///
/// @brief      Get temperature threshold
/// @param      in_strDeviceName is a string for temperature threshold
/// @param      out_fpValue is a pointer to a FLOAT for the threshold
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::getTemperatureThreshold(string   in_strDeviceName,
                                  FLOAT    *out_fpValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    INT32       s32RegOffset = 0x0, s32Count = 0;
    UINT16      u16Data;

    if (OT_WARN_LIMIT_NAME && ((CHAR *)OT_WARN_LIMIT_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_OT_WARN_LIMIT;
    }
    else if (OT_FAULT_LIMIT_NAME && ((CHAR *)OT_FAULT_LIMIT_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_OT_FAULT_LIMIT;
    }
    else if (UT_WARN_LIMIT_NAME && ((CHAR *)UT_WARN_LIMIT_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_UT_WARN_LIMIT;
    }
    else if (UT_FAULT_LIMIT_NAME && ((CHAR *)UT_FAULT_LIMIT_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_UT_FAULT_LIMIT;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_GET_TEMPERATURE_THRESHOLD | ERROR_NO_DEVICE_FOUND),
                              in_strDeviceName.c_str());
    }

    pErrorMsg = m_pI2CDAPI->Read(getName(), s32RegOffset, s32RegOffset, &u16Data, &s32Count);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_GET_TEMPERATURE_THRESHOLD, NULL, pErrorMsg);
    }

    *out_fpValue = transformLinearToHuman(u16Data);

    return NO_ERROR_MSG;
}
/// @fn         PMBusPSU::getTemperatureThresholdList(list<string> *out_pList)
///
/// @brief      Get the list of temperature sensor thresholds
/// @param      out_pList is a pointer to a list of strings for temperature sensor names
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::getTemperatureThresholdList(list<string> *out_pList)
{
    if (OT_WARN_LIMIT_NAME)
    {
        out_pList->push_back(OT_WARN_LIMIT_NAME);
    }

    if (OT_FAULT_LIMIT_NAME)
    {
        out_pList->push_back(OT_FAULT_LIMIT_NAME);
    }

    if (UT_WARN_LIMIT_NAME)
    {
        out_pList->push_back(UT_WARN_LIMIT_NAME);
    }

    if (UT_FAULT_LIMIT_NAME)
    {
        out_pList->push_back(UT_FAULT_LIMIT_NAME);
    }

    return NO_ERROR_MSG;
}

/// @fn         PMBusPSU::getType(string *out_strpValue)
///
/// @brief      Get protocol type
/// @param      out_strpValue is a pointer to a string for the protocol type
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::getType(string *out_strpValue)
{
    *out_strpValue = "PMBus";

    return NO_ERROR_MSG;
}

/// @fn         PMBusPSU::isPresent(BOOL *out_bpPresent)
///
/// @brief      Check if PSU is present
/// @param      out_bpPresent is a pointer to a BOOL for the status
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::isPresent(BOOL *out_bpPresent)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT8       u8Data = 0;
    INT32       s32Count = 0;

    pErrorMsg = m_pI2CDAPI->Read(getName(),
                                 PMBUS_CMD_PMBUS_REVISION,
                                 PMBUS_CMD_PMBUS_REVISION,
                                 &u8Data,
                                 &s32Count);

    if (pErrorMsg != NO_ERROR_MSG)
    {
        deleteErrorMsg(pErrorMsg);
        *out_bpPresent = FALSE;
        return NO_ERROR_MSG;
    }

    if ((u8Data == 0x00) || (u8Data == 0x11) || (u8Data == 0x22))
    {
        *out_bpPresent = TRUE;
        return NO_ERROR_MSG;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_IS_PRESENT | ERROR_UNEXPECTED_VALUE));
    }
}

/// @fn         PMBusPSU::setFanSpeed(string   in_strDeviceName,
///                                   INT32    in_s32Value)
/// @brief      Set FAN speed to the given FAN
/// @param      in_strDeviceName is a string for the FAN
/// @param      in_s32Value is a INT32 for the FAN speed
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::setFanSpeed(string   in_strDeviceName,
                      INT32    in_s32Value)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    INT32       s32RegOffset = 0x0;
    UINT16      u16Data;

    if (FAN1_NAME && ((CHAR *)FAN1_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_FAN_COMMAND_1;
    }
    else if (FAN2_NAME && ((CHAR *)FAN2_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_FAN_COMMAND_2;
    }
    else if (FAN3_NAME && ((CHAR *)FAN3_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_FAN_COMMAND_3;
    }
    else if (FAN4_NAME && ((CHAR *)FAN4_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_FAN_COMMAND_4;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_SET_FAN_SPEED | ERROR_NO_DEVICE_FOUND),
                              in_strDeviceName.c_str());
    }

    u16Data = transformHumanToLinear((FLOAT)in_s32Value);

    pErrorMsg = m_pI2CDAPI->Write(getName(), s32RegOffset, &u16Data, 1);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_SET_FAN_SPEED, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

/// @fn         PMBusPSU::setTemperatureThreshold(string   in_strThreshold,
///                                               FLOAT    in_fValue)
/// @brief      Set temperature threshold for temperature sensors
/// @param      in_strDeviceName is a string for the temperature threshold name
/// @param      in_fValue is a FLOAT for the threshold value
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
PMBusPSU::setTemperatureThreshold(string   in_strDeviceName,
                                  FLOAT    in_fValue)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    INT32       s32RegOffset = 0x0, s32Count = 0;
    UINT16      u16Data;

    if (OT_WARN_LIMIT_NAME && ((CHAR *)OT_WARN_LIMIT_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_OT_WARN_LIMIT;
    }
    else if (OT_FAULT_LIMIT_NAME && ((CHAR *)OT_FAULT_LIMIT_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_OT_FAULT_LIMIT;
    }
    else if (UT_WARN_LIMIT_NAME && ((CHAR *)UT_WARN_LIMIT_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_UT_WARN_LIMIT;
    }
    else if (UT_FAULT_LIMIT_NAME && ((CHAR *)UT_FAULT_LIMIT_NAME == in_strDeviceName))
    {
        s32RegOffset = PMBUS_CMD_UT_FAULT_LIMIT;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_PSU_DDAPI_SET_TEMPERATURE_THRESHOLD | ERROR_NO_DEVICE_FOUND),
                              in_strDeviceName.c_str());
    }

    u16Data = transformHumanToLinear(in_fValue);

    pErrorMsg = m_pI2CDAPI->Write(getName(), s32RegOffset, &u16Data, 1);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_PSU_DDAPI_SET_TEMPERATURE_THRESHOLD, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}

UINT16
PMBusPSU::transformHumanToLinear(FLOAT  in_fValue)
{
    FLOAT   fCurrentDifference = 0.0;
    FLOAT   fTempDifference = 0.0;
    FLOAT   fTemp = 0.0;
    INT32   s32Y = 0;
    INT32   i = 0;
    UINT16  u16LinearValue = 0;

    if (in_fValue >= (FLOAT)0)
    {
        fCurrentDifference = in_fValue;

        // Consider that the exponent N >= 0
        if (in_fValue >= (FLOAT)1.0)
        {
            for (i = 0; i < 16; i++)
            {
                fTemp = (FLOAT)(1 << i);

                if (fTemp <= in_fValue)
                {
                    s32Y = (INT32)(in_fValue / fTemp);

                    if (s32Y > 1023)
                    {
                        continue;
                    }
                    else
                    {
                        fTemp = (FLOAT)(s32Y << i);

                        // If new value is more closing to the expected value, calculate new linear value
                        if (fTemp <= in_fValue)
                        {
                            fTempDifference = in_fValue - fTemp;
                        }
                        else
                        {
                            fTempDifference = fTemp - in_fValue;
                        }

                        if (fTempDifference < fCurrentDifference)
                        {
                            u16LinearValue = LINEAR_VALUE(s32Y, i);
                            fCurrentDifference = fTempDifference;

                            if (fCurrentDifference == 0)
                            {
                                break;
                            }
                        }
                    }
                }
                else
                {
                    break;
                }
            }
        }
        // Consider that the exponent N < 0
        else
        {
            for (i = 15; i > 0; i--)
            {
                fTemp = (1.0 / (FLOAT)(1 << i));

                if (fTemp <= in_fValue)
                {
                    s32Y = (INT32)(in_fValue * (1 << i));

                    if ((s32Y > 1023) || (s32Y == 0))
                    {
                        continue;
                    }
                    else
                    {
                        fTemp = (FLOAT)s32Y * fTemp;

                        // If new value is more closing to the expected value, calculate new linear value
                        if (fTemp <= in_fValue)
                        {
                            fTempDifference = in_fValue - fTemp;
                        }
                        else
                        {
                            fTempDifference = fTemp - in_fValue;
                        }

                        if (fTempDifference < fCurrentDifference)
                        {
                            u16LinearValue = LINEAR_VALUE(s32Y, -i);
                            fCurrentDifference = fTempDifference;
                        }
                    }
                }
                else
                {
                    break;
                }
            }
        }
    }
    else
    {
        fCurrentDifference = -in_fValue;

        // Consider that the exponent N >= 0
        if (in_fValue <= (FLOAT)-1.0)
        {
            for (i = 0; i < 16; i++)
            {
                fTemp = (FLOAT)(1 << i);

                if (fTemp <= -in_fValue)
                {
                    s32Y = (INT32)(in_fValue / fTemp);

                    if (s32Y < -1024)
                    {
                        continue;
                    }
                    else
                    {
                        fTemp = (FLOAT)(s32Y << i);

                        // If new value is more closing to the expected value, calculate new linear value
                        if (fTemp <= in_fValue)
                        {
                            fTempDifference = in_fValue - fTemp;
                        }
                        else
                        {
                            fTempDifference = fTemp - in_fValue;
                        }

                        if (fTempDifference < fCurrentDifference)
                        {
                            u16LinearValue = LINEAR_VALUE(s32Y, i);
                            fCurrentDifference = fTempDifference;
                        }
                    }
                }
                else
                {
                    break;
                }
            }
        }
        // Consider that the exponent N < 0
        else
        {
            for (i = 15; i > 0; i--)
            {
                fTemp = (1.0 / (FLOAT)(1 << i));

                if (fTemp <= -in_fValue)
                {
                    s32Y = (INT32)(in_fValue * (1 << i));

                    if ((s32Y < -1024) || (s32Y == 0))
                    {
                        continue;
                    }
                    else
                    {
                        fTemp = (FLOAT)s32Y * fTemp;

                        // If new value is more closing to the expected value, calculate new linear value
                        if (fTemp <= in_fValue)
                        {
                            fTempDifference = in_fValue - fTemp;
                        }
                        else
                        {
                            fTempDifference = fTemp - in_fValue;
                        }

                        if (fTempDifference < fCurrentDifference)
                        {
                            u16LinearValue = LINEAR_VALUE(s32Y, -i);
                            fCurrentDifference = fTempDifference;
                        }
                    }
                }
                else
                {
                    break;
                }
            }
        }
    }

    return u16LinearValue;
}

FLOAT
PMBusPSU::transformLinearToHuman(UINT16 in_u16Data)
{
    INT32   s32Y = Y_VALUE(in_u16Data);
    INT32   s32N = N_VALUE(in_u16Data);
    FLOAT   fValue = (FLOAT)s32Y;

    return (s32N >= 0)? (fValue * (1 << s32N)) : (fValue / (1 << -s32N));
}
